

"""
This module defines a number of functions and classes used throughout
the program.
"""

import os, copy, collections

# Configuration defaults

_CFG = {
"version"         : "3.5.9",             # The skipole version
"default_language": 'en',                # The default language of the project
"debug"           : False                # The debug mode, True shows exceptions on server error
}

# create a dictionary of all projects, {proj_ident:project}
PROJECT_REGISTER = {}

_LIB_LABELS = ["skipole_js", "jquery_core",
               "ski_basic", "ski_checkbox", "ski_confirm", "ski_debug_tools", "ski_dropdown", "ski_error_messages",
               "ski_footers", "ski_headers", "ski_info", "ski_inputforms", "ski_inputtables",
               "ski_inputtext", "ski_links", "ski_lists", "ski_logins", "ski_paras", "ski_paths", "ski_radio", "ski_svgbasics", "ski_svggraphs", "ski_svgmeters",
               "ski_tables", "ski_textarea", "ski_upload"]

_SYS_LABELS = ["url_not_found", "validate_error", "server_error", "redirector", "no_javascript", "general_json"]


Info = collections.namedtuple('Info', ['project', 'project_version', 'ident', 'item_type', 'name', 'brief', 'path', 'label_list', 'change', 'parentfolder_ident', 'restricted'])


def add_to_project_register(project):
    "Adds the project to a list of all projects"
    global PROJECT_REGISTER
    PROJECT_REGISTER[project.proj_ident] = project


def project_register():
    "Return a dictionary of all projects"
    global PROJECT_REGISTER
    return PROJECT_REGISTER

def project_ident_register():
    "Return a list of all project idents"
    global PROJECT_REGISTER
    return list(PROJECT_REGISTER.keys())

def root_project():
    "Return the root project"
    global PROJECT_REGISTER
    for proj in PROJECT_REGISTER.values():
        if proj.rootproject:
            return proj

def projectfiles(proj_ident=None):
    """Returns the directory entry where projectfiles can be found for a given project
       If proj_ident is None, then returns projectfiles for the root project"""
    proj = getproject(proj_ident)
    if proj is None:
        return
    return proj.projectfiles

def lib_list():
    "Returns list of library labels"
    return _LIB_LABELS

def sys_list():
    "Returns list of system page labels"
    return _SYS_LABELS

def is_project(proj_ident):
    "Returns True if this project is in the site, False otherwise"
    if proj_ident in PROJECT_REGISTER:
        return True
    return False

def get_debug():
    "Returns the debug mode"
    return _CFG["debug"]

def set_debug(mode):
    "Sets debug mode"
    global _CFG
    _CFG["debug"] = bool(mode)


def getproject(proj_ident):
    """Returns the project given by the proj_ident
       If the project does not exist, return None"""
    if proj_ident in PROJECT_REGISTER:
        return PROJECT_REGISTER[proj_ident]

def project_ident(proj_ident=None):
    "Returns the given project ident, if it is None, returns current site root project ident"
    if proj_ident is None:
        return root_project().proj_ident
    return proj_ident

def projectdir(proj_ident=None):
    "Returns projectfiles/proj_ident"
    return os.path.join(projectfiles(proj_ident), project_ident(proj_ident))

def projectstatic(proj_ident=None):
    "Returns projectfiles/proj_ident/static"
    return os.path.join(projectdir(proj_ident), 'static')

def projectdata(proj_ident=None):
    "Returns projectfiles/proj_ident/data"
    return os.path.join(projectdir(proj_ident), 'data')

def projectcode(proj_ident=None):
    "Returns projectfiles/proj_ident/code"
    return os.path.join(projectdir(proj_ident), 'code')


def project_json(proj_ident=None):
    "Returns projectfiles/proj_ident/data/project.json"
    return os.path.join(projectdata(proj_ident), "project.json")

def project_defaults(proj_ident=None):
    "Returns projectfiles/proj_ident/data/defaults.json"
    return os.path.join(projectdata(proj_ident), "defaults.json")

def default_language():
    "Returns the default language"
    return _CFG["default_language"]

def version():
    "Returns the version string"
    return _CFG["version"]

    

# These functions deal with page and folder 'idents'.  They are defined here
# as they are common functions used everywhere.

def from_ident(ident, proj_ident=None):
    """given an Ident, or a string version of ident, or item with ident attribute, return deepcopy of page or folder."""
    ident = Ident.to_ident(ident, proj_ident)
    if ident is None:
        return
    project = getproject(ident.proj)
    if project is None:
        return
    item = project.get_item(ident)
    return copy.deepcopy(item)


def item_info(ident):
    """Returns a namedtuple of item information
          None if not found, tuple has contents:
          project, project_version, ident, item_type, name, brief, path, label_list, change, parentfolder_ident, restricted
       note: project is the project name
             parentfolder_ident is the parent folder ident or None if this is root
             ident is this item ident"""

    project = getproject(ident.proj)
    if project is None:
        return
    item = project.get_item(ident)
    if item is None:
        return

    if item.name is None:
        name = ''
    else:
        name = item.name
    special_pages = project.special_pages
    label_list = []
    for label, value in special_pages.items():
        if ident == value:
            label_list.append(label)
    return Info(ident.proj,
                project.version,
                ident,
                item.page_type,
                name,
                item.brief,
                item.url,
                label_list,
                item.change,
                item.parentfolder_ident,
                item.restricted
               )


def get_item(ident):
    """Gets uncopied page or folder, without importing sections and returns it
        Returns None if not found. Requires full ident object"""
    project = getproject(ident.proj)
    if project is None:
        return None
    return project.get_item(ident)


def get_url(label_url_ident, proj_ident=None):
    "Returns a url given a page ident, url or label, if not found, return None"
    if (label_url_ident is not 0) and (not label_url_ident):
        return
    result = find_ident_or_url(label_url_ident, proj_ident)
    if isinstance(result, Ident):
        return result.url()
    return result


def ident_from_path(path, proj_ident=None):
    project = getproject(proj_ident)
    if project is None:
        return
    return project.root.ident_from_path(path)


def ident_exists(ident, proj_ident=None):
    "Return True if ident exists, False otherwise, ident can be a string form of ident"
    ident = Ident.to_ident(ident, proj_ident)
    if ident is None:
        return False
    project = getproject(ident.proj)
    if project is None:
        return False
    return ident.num in project.ident_numbers


def ident_exists_strict(ident):
    "Returns ident if it exists, None otherwise, ident must be an Ident object"
    if not isinstance(ident, Ident):
        return
    project = getproject(ident.proj)
    if project is None:
        return
    if ident.num in project.ident_numbers:
        return ident


# usefull functions
    

def mergedict(dict1, dict2):
    "returns a dictionary, which is the update of two dictionaries"
    newdict = dict1.copy()
    newdict.update(dict2)
    return newdict

def mergedicttostyle(style_dict, key, add_dict):
    "Returns a new style_dict, with the dictionary of the key item updated with add_dict"
    if key in style_dict:
        new_sub_dict = mergedict(style_dict[key], add_dict)
    else:
        new_sub_dict = add_dict.copy()
    new_dict = style_dict.copy()
    new_dict[key] = new_sub_dict
    return new_dict


def make_ident_or_label_or_url(item, proj_ident=None):
    """Given an item, be it integer, string or Ident, return an Ident object
       If the item is an integer, the returned Ident object
       will have its proj set to the proj_ident if given, otherwise current project
       If item is a not convertable to an ident - then return the string,
       which should be a label or url"""
    if isinstance(item, str) and ('/' in item):
        # must be a url
        return item
    ident = Ident.to_ident(item, proj_ident)
    if ident is None:
        # not convertable to ident, must be a label
        return item
    return ident


def make_ident_or_label(item, proj_ident=None):
    """Given an item, be it integer, string or Ident, return an Ident object
       or label - not a url.
       If the item is an integer, the returned Ident object
       will have its proj set to the proj_ident if given, otherwise current project
       If item is a not convertable to an ident - then return the string,
       which should be a label, if string contains /, return None"""
    if isinstance(item, str) and ('/' in item):
        # a url - return None
        return
    ident = Ident.to_ident(item, proj_ident)
    if ident is None:
        # not convertable to ident, must be a label
        return item
    return ident



def make_ident(item, proj_ident=None):
    """Just calls Ident.to_ident"""
    return Ident.to_ident(item, proj_ident)


def find_ident(item, proj_ident=None):
    """Given any item makes an attempt to find an Ident
       existing in the project, item could have an ident attribute,
       or be an ident in string or integer or Ident form
       or label, or tuple of form (project,ident) or (project,label).
       If no existing ident or if a url found instead, returns None"""

    ident_or_url = find_ident_or_url(item, proj_ident)
    if isinstance(ident_or_url, Ident):
        return ident_or_url
    # ident_or_url may be a URL or None, in this case return None
    return



def find_ident_or_url(item, proj_ident=None):
    """Given any item makes an attempt to find an Ident or url
       existing in the project, item could have an ident attribute,
       or be an ident in string or integer or Ident form
       or label, or tuple of form (project,ident) or (project,label).
       If no existing ident or url found, returns None"""

    if item is 0:
        return ident_exists_strict(Ident.to_ident(item, proj_ident))

    if not item:
        return

    if isinstance(item, Ident):
        return ident_exists_strict(item)

    if isinstance(item, str):
        item = item.strip()
        if not item:
            return
        if '/' in item:
            # its a url
            return item
        if item.isdigit():
            # string integer
            return ident_exists_strict(Ident.to_ident(item, proj_ident))
        # could be 'project,number' or 'project,label', 
        if ',' in item:
            item_parts = item.split(',')
            if len(item_parts) != 2:
                return
            if item_parts[1].isdigit():
                return ident_exists_strict(Ident.to_ident(item, proj_ident))
            # may be 'project,label' - so find [project,label]
            return find_ident_or_url(item_parts, proj_ident=proj_ident)
        # could be case of 'project_number'
        if '_' in item:
            item_parts = item.split('_')
            if (len(item_parts) == 2) and item_parts[1].isdigit():
                return ident_exists_strict(Ident.to_ident(item, proj_ident))

        # item is a label
        project = getproject(proj_ident)
        if project is None:
            return
        special_pages = project.special_pages
        if item not in special_pages:
            return
        target = special_pages[item]
        # target is either an Ident, url, or 'proj,label' string
        if isinstance(target, Ident):
            return ident_exists_strict(target)
        if '/' in target:
            # so item is a label pointing to this url
            return target
        if ',' in target:
            # so item is a label pointing to project,newitem
            target_parts = target.split(',')
            return find_ident_or_url(target_parts[1], proj_ident=target_parts[0])

    if not ( (isinstance(item, tuple)) or (isinstance(item, list)) ):
        # not a tuple or a list, so maybe integer - try to convert to ident
        return ident_exists_strict(Ident.to_ident(item, proj_ident))

    # test for [project, newitem]

    # item is a tuple or list
    if len(item) != 2:
        # invalid, return None
        return

    proj = item[0].strip()

    project = getproject(proj)
    if project is None:
        return

    return find_ident_or_url(item[1], proj_ident=proj)




# Each folder and page has an instance of Ident
class Ident(collections.namedtuple('Ident', ['proj','num'])):
    __slots__ = ()

    @classmethod
    def to_ident(cls, item, proj_ident=None):
        """Given an ident, be it tuple, integer, string or Ident, return an Ident object
           If the item is just an integer, the returned Ident object
           will have its proj set to the proj_ident if given, otherwise current project
           If an Ident object cannot be created, return None"""
        # Deal with cases where proj_ident is not needed
        if isinstance(item, cls):
            return item
        if isinstance(item, str):
            # first deal with case of a string ident myproj_num
            if '_' in item:
                item = item.split('_')
            elif ',' in item:
                item = item.split(',')
            else:
                # could be a string integer
                try:
                    item = int(item)
                except Exception:
                    # don't know what it is, return None
                    return
        if isinstance(item, tuple) or isinstance(item, list):
            if len(item) == 2:
                try:
                    num = int(item[1])
                except Exception:
                    # don't know what it is, return None
                    return
                if item[0] and isinstance(item[0], str):
                    return cls(item[0], num)
                else:
                    return
            else:
                # list, but not of length two, don't know what it is
                return
        if hasattr(item, 'ident') and (isinstance(item.ident, cls)):
            return item.ident
        if isinstance(item, int):
            # project not derived from item, therefore need to use proj_ident
            # proj_ident must be either an Ident instance, string or None
            if isinstance(proj_ident, cls):
                proj = proj_ident.proj
            elif proj_ident:
                proj = proj_ident
            else:
                proj = root_project().proj_ident
            return cls(proj, item)

    def item(self):
        """Return page or folder with this ident. 
           If folder or respond page return the item, any other page, return a deep copy
           If not found, return None."""
        project = getproject(self.proj)
        if project is None:
            return
        try:
            item = project[self]
        except Exception:
            return
        return item


    def name(self):
        """Returns the name of the page or folder with this ident
           If item is root, or item not found, returns None"""
        if self.num is 0:
            return None
        item = get_item(self)
        if item is None:
            return
        return item.name

    def url(self):
        "Returns the url of the page or folder with this ident"
        item = get_item(self)
        if item is None:
            return
        return item.url

    def widg_ident(self, widgname):
        """Given a widget name, if this ident is the ident of a page, and the widget
           is in the page, return the widgets ident, if the ident is not that of a page
           or the widget name is not found in the page, return an empty string"""
        item = get_item(self)
        if ((item.page_type == 'TemplatePage') or (item.page_type == 'SVG')) and (widgname in item.widgets):
            return item.widgets[widgname]
        return ''

    def to_tuple(self):
        return (self.proj, self.num)

    def to_comma_str(self):
        return self.proj + "," + str(self.num)

    def __str__(self):
        return self.proj + "_" + str(self.num)

    def __add__(self, other):
        "An ident can be added to a integer or tuple to give a string"
        if isinstance(other, Ident):
            raise TypeError("An Ident cannot be added to another Ident")
        if isinstance(other, tuple) or isinstance(other, list):
            return self.__str__() + '_' + '_'.join(str(i) for i in other)
        return self.__str__() + '_' + str(other)

#
# Defines functions and class WidgField which is a named tuple with attributes s,w,f,i
# for section name, widget name, field name, index value
# section name could be empty if widget is not in a section
# index value is used for widgfield dictionaries, and can be empty if not applicable
# widgfield has string sectionname-widgetname:fieldname-index
#

def make_widgfield(field, widgetonly=False):
    """Given a string field, of the form sectionname-widgetname:fieldname-index, (set by widgets)
       or of the form sectionname-widgetname:fieldname.index, (set by browser, ie image x,y coordinates)
       or a list or tuple, of form (sectionname, widgetname, fieldname, index)
       Note: if index given, tuple must have four elements, if three given they will
       be assumed as (sectionname, widgetname, fieldname)
       returns a named tuple of the form WidgField(s=sectionname, w=widgetname, f=fieldname, i=index)
       If field is already a WidgField, just returns it"""
    if not field:
        return WidgField(s='', w='', f='', i='')
    if widgetonly:
        if isinstance(field, WidgField):
            return WidgField(s=field.s, w=field.w, f='', i='')
        if isinstance(field, str) and (',' in field):
            field = [ item.strip() for item in field.split(',') ]
        if isinstance(field, list) or isinstance(field, tuple):
            if len(field) == 1:
                return WidgField(s='', w=field[0], f='', i='')
            if len(field) >= 2:
                return WidgField(s=field[0], w=field[1], f='', i='')
        # so should be a string
        if ':' in field:
            field = field.split(':')[0]
        if not field:
            return WidgField(s='', w='', f='', i='')
        if '-' in field:
            sectionname, widgetname = field.split("-", 1)
            return WidgField(s=sectionname, w=widgetname, f='', i='')
        else:
            # widgetname only
            return WidgField(s='', w=field, f='', i='')
    # not just a widget, could be widget and field
    if isinstance(field, WidgField):
        return field
    if isinstance(field, str) and (',' in field):
        field = [ item.strip() for item in field.split(',') ]
    if isinstance(field, list) or isinstance(field, tuple):
        if len(field) == 1:
            return WidgField(s='', w=field[0], f='', i='')
        if len(field) == 2:
            return WidgField(s='', w=field[0], f=field[1], i='')
        if len(field) == 3:
            return WidgField(s=field[0], w=field[1], f=field[2], i='')
        return WidgField(s=field[0], w=field[1], f=field[2], i=field[3])
    # so should be a string
    if ':' not in field:
        if '-' in field:
            sectionname, widgetname = field.split("-", 1)
            return WidgField(s=sectionname, w=widgetname, f='', i='')
        else:
            # widgetname only, no section, field or index
            return WidgField(s='', w=field, f='', i='')
    sectionwidgetname, field_index = field.split(":", 1)
    if '-' in sectionwidgetname:
        sectionname, widgetname = sectionwidgetname.split("-", 1)
    else:
        sectionname = ''
        widgetname = sectionwidgetname
    # field index separated by - or by .
    if '-' in field_index:
        fieldname, index = field_index.split("-", 1)
    elif '.' in field_index:
        fieldname, index = field_index.split(".", 1)
    else:
        fieldname = field_index
        index=''
    return WidgField(s=sectionname, w=widgetname, f=fieldname, i=index)


def list_widgfield(widgfield, widgfield_dict):
    """Converts a widgfield (with no i value), and a dictionary to a list of
       of widgfield,value tuples, with each widgfield in the tuple having i set to the key
       and the list is sorted by key"""
    if widgfield.i:
        return [(widgfield, widgfield_dict)]
    return sorted([(widgfield._replace(i=key), val) for key, val in widgfield_dict.items()], key=lambda tup: tup[0].i)


class WidgField(collections.namedtuple('WidgField', 's w f i')):
    __slots__ = ()

    def set_field(self, f='', i=''):
        "Returns a copy of self, with new field and index set"
        return WidgField(s=self.s, w=self.w, f=f, i=i)

    def set_index(self, i=''):
        "Returns a copy of self, with new index set"
        return WidgField(s=self.s, w=self.w, f=self.f, i=i)

    def __bool__(self):
        if self.s or self.w or self.f or self.i:
            return True
        return False

    def __str__(self):
        "returns sectionname-widgetname:fieldname-index"
        if self.s:
            val = self.s + '-' + self.w
        else:
            val = self.w
        if self.f:
             val = val + ':' + self.f
        if self.i:
            val = val + '-' + self.i
        return val

    def to_tuple(self):
        "returns tuple of (self.s, self.w, self.f, self.i)"
        return (self.s, self.w, self.f, self.i)

    def to_tuple_no_i(self):
        "returns tuple of (self.s, self.w, self.f) if self.s or (self.w, self.f) if no self.s"
        if self.s:
            return (self.s, self.w, self.f)
        return (self.w, self.f)

    def sw_tuple(self):
        "returns tuple of (self.s, self.w)"
        if self.s:
            return (self.s, self.w)
        return ('',self.w)

    def to_str_tuple(self):
        "returns sectionname,widgetname,fieldname,index"
        if self.s:
            val = self.s + ',' + self.w
        else:
            val = self.w
        if self.f:
             val = val + ',' + self.f
        if self.i:
            val = val + ',' + self.i
        return val








"""This package contains modules of Widget classes which can be added to pages.

These Widgets inherit from the classes defined here, which themselves inherit
from tag.Part and tag.ClosedPart"""

import html, copy, collections, re, json, datetime
from string import Template

from .. import tag, skiboot
from ..excepts import ServerError, ValidateError

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')

#########################################################
# Field Arguments consist of a set of classes which
# act as the input for Widgets
#########################################################


# Each of these field arguments carries a value, which can be one
# of the following types:

# text
# url           - either a url string, label string or ident
# textblock_ref
# ident         - the ident of a page or folder
# element_ident - the string ident of a page element
# boolean
# integer
# widgfield
# cssclass
# cssstyle
# date          - python datetime.date object
# datetime      - python datetime.datetime object


class ParentFieldArg(object):
    "The parent class of FieldArg classes"

    def __init__(self, field_type, valdt=False, jsonset=False, senddict = False):
        """Acts as a parent to the field argument classes"""
        self.field_type = field_type
        # the name is usually set to the argument name when a widget is initialised,
        # but can be altered
        self.name = None
        self._value = None
        # val_list is a list of validators
        self.val_list = []
        # this is true if the expected value submitted by the browser is a dictionary
        self.senddict = senddict
        # True if validators can be added
        self.valdt = valdt
        # True if this field can be set by json
        self.jsonset = jsonset
        # True if field has a default css class or style, only valid for FieldArg
        self.cssclass = False
        self.cssstyle = False

    def _typematch(self, val, valtype):
        "Returns val, matched to valtype"
        try:
            if valtype == 'boolean':
                if isinstance(val, str):
                    if val.lower() == 'false':
                        return False
                return bool(val)
            elif valtype == 'ident':
                if val is 0:
                    return skiboot.make_ident(val)
                if not val:
                    return None
                return skiboot.make_ident(val)
            elif valtype == 'url':
                if val is 0:
                    sval = '0'
                else:
                    if not val:
                        return ''
                    sval = str(val)
                if '/' in sval:
                    # a url
                    return sval
                # could be an ident or label
                if ',' in sval:
                    labelparts = sval.split(',')
                    if labelparts[0].isalnum() and labelparts[1].isnumeric():
                        # probably an ident
                        ident = skiboot.make_ident(sval)
                        if ident is None:
                            return ''
                        return ident
                if '_' in sval:
                    # _ exists, but could still be a label
                    labelparts = sval.split('_')
                    if labelparts[0].isalnum() and labelparts[1].isnumeric():
                        # probably an ident
                        ident = skiboot.make_ident(sval)
                        if ident is not None:
                            return ident
                # a label
                return sval
            elif valtype == 'textblock_ref':
                if not val:
                    val = tag.TextBlock('')
                elif not isinstance(val, tag.TextBlock):
                    val = tag.TextBlock(str(val))
                # so val is a tag.TextBlock, set its parameters the same as self._value
                if isinstance(self._value, tag.TextBlock):
                    if (not val.failmessage_set()) and self._value.failmessage_set():
                        val.failmessage = self._value.failmessage
                    val.linebreaks = self._value.linebreaks
                    val.project = self._value.project
            elif valtype == 'integer':
                if not val:
                    return 0
                try:
                    int_val = int(val)
                except Exception:
                    raise ValidateError("Given value invalid, should be integer")
                return int_val
            elif valtype =='widgfield':
                if not val:
                    return ''
                return skiboot.make_widgfield(val)
            elif (valtype == 'text') or (valtype == 'cssclass') or (valtype == 'cssstyle'):
                return str(val)
            elif valtype =='date':
                if isinstance(val, datetime.date):
                    return val
                elif not val:
                    return ''
                elif isinstance(val, str):
                    # only accept yyyy-mm-dd
                    try:
                        yearstring,monthstring,daystring = val.split('-')
                        year = int(yearstring)
                        month = int(monthstring)
                        day = int(daystring)
                        thisday = datetime.date(year,month,day)
                    except:
                        raise ValidateError("Given value invalid, should be yyyy-mm-dd or a datetime.date object")
                    return thisday
                else:
                    raise ValidateError("Given value invalid, should be a datetime.date object")
            elif valtype =='datetime':
                if not val:
                    return datetime.datetime.utcnow()
                elif isinstance(val, datetime.datetime):
                    return val
                else:
                    raise ValidateError("Given value invalid, should be a datetime.datetime object")
            else:
                raise ValidateError("Field type %s not recognised" % (valtype,))
        except Exception as e:
            if hasattr(e, 'message'):
                if e.message:
                    raise ValidateError(e.message)
            raise ValidateError("Given value invalid")
        return val

    @property
    def string_value(self):
        "The string version of the field value"
        return ''

    def __len__(self):
        return 1

    def __bool__(self):
        return bool(self._value)

    def add_validator(self, val):
        if self.valdt:
            self.val_list.append(val)

    def validate(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        """validates against the val_list returns item, empty list on success
           or empty item, errorlist on failure - however
           If item is a dictionary, successful items will have value, failed items will be empty
           If item is a list, any item failing will produce an empty item"""
        if not self.valdt:
            raise ValidateError('Invalid field to apply validator')
        if not self.val_list:
            if widgfield:
                message = "No validator has been set on field %s" % (widgfield,)
            elif self.name:
                message = "No validator has been set on field %s" % (self.name,)
            else:
                message = "No validator has been set on field"
            raise ValidateError(message)
        if isinstance(item, dict):
            error_list = []
            newdict = item.copy()
            # need to test each item in newdict against val list - but some items may pass
            # others may fail, so have to keep testing the passed ones against further tests
            for val in self.val_list:
                if not newdict:
                    break
                newdict, errors = val(widgfield, newdict, environ, lang, form_data, call_data, caller_page_ident)

                if errors:
                    error_list.extend(errors)
            # newdict is a dictionary of ok values, all other values in item are set to empty string
            for key, val in item.items():
                if key in newdict:
                    item[key] = newdict[key]
                else:
                    item[key] = ''
        else:
            for val in self.val_list:
                item, error_list = val(widgfield, item, environ, lang, form_data, call_data, caller_page_ident)
                if error_list:
                    break
        return item, error_list


class FieldArg(ParentFieldArg):
    "A basic single value field"

    def __init__(self, field_type, default='', valdt=False, jsonset=False, senddict = False):
        """
        default is the initial value of this field
        """
        ParentFieldArg.__init__(self, field_type, valdt=valdt, jsonset=jsonset, senddict=senddict)
        self.value = default
        # cssclass is True if this field represents a css class,
        # and can be set with defaults from defaults.json
        if field_type == 'cssclass':
            self.cssclass = True
        else:
            self.cssclass = False
        if field_type == 'cssstyle':
            self.cssstyle = True
        else:
            self.cssstyle = False

    def get_value(self):
        return self._value

    def set_value(self, val):
        self._value = self._typematch(val, self.field_type)

    value = property(get_value, set_value)

    @property
    def string_value(self):
        "Converts idents and WidgFields to strings, Textblocks to textref's"
        if self.field_type == 'textblock_ref':
            return self._value.textref
        if self.field_type == 'boolean':
            if self._value:
                return "True"
            else:
                return "False"
        if self._value is None:
            return ''
        return str(self._value)



class FieldArgList(ParentFieldArg):
    "A field which takes a list of items"

    def __init__(self, field_type, valdt=False, jsonset=False, senddict = False):
        """A list
        """
        ParentFieldArg.__init__(self, field_type, valdt=valdt, jsonset=jsonset, senddict=senddict)
        self._value = []

    def get_value(self):
        return self._value

    def set_value(self, val):
        # val should be a list
        self._value = []
        if not val:
            return
        elif (not isinstance(val, list)) and (not isinstance(val, tuple)):
            val = [val]
        for item in val:
            self._value.append(self._typematch(item, self.field_type))

    value = property(get_value, set_value)

    @property
    def string_value(self):
        "Converts list of idents and WidgFields to list of strings, Textblocks to textref's"
        value_list = []
        for val in self._value:
            if self.field_type == 'textblock_ref':
                value_list.append(val.textref)
            elif self.field_type == 'boolean':
                value_list.append(val)
            elif val is None:
                value_list.append('')
            else:
                value_list.append(str(val))
        return value_list

    def __len__(self):
        return len(self._value)

    def __bool__(self):
        "Returns boolean value of the list"
        return bool(self._value)


class FieldArgTable(ParentFieldArg):
    """A field which takes a table of items - a list of rows, each row being a list
       for example: [[r1c1, r1c2, r1c3], [r2c1, r2c2, r2c3], [r3c1, r3c2, r3c3]]
    """

    def __init__(self, field_type, valdt=False, jsonset=False, senddict = False):
        """
           field_type is a list or tuple of column types
        """
        ParentFieldArg.__init__(self, field_type, valdt=valdt, jsonset=jsonset, senddict=senddict)
        self.value = None


    @property
    def cols(self):
        return len(self.field_type)

    def get_value(self):
        return self._value

    def set_value(self, val):
        "Sets the value"
        # val should be a list of row lists
        # ie [[r1c1, r1c2, r1c3], [r2c1, r2c2, r2c3], [r3c1, r3c2, r3c3]]
        self._value = []
        if val is 0:
             val = [[0]]
        elif not val:
            return
        elif (not isinstance(val, list)) and (not isinstance(val, tuple)):
            val = [[val]] 
        cols = len(self.field_type)
        for row in val:
            if (not isinstance(row, list)) and (not isinstance(row, tuple)):
                row = [row]
            reqrow = []         # required row
            for idx, col in enumerate(row):
                if idx >= cols:
                    continue
                reqrow.append(self._typematch(col, self.field_type[idx]))
            # If the reqrow does not have enough column entries, pad it out
            if len(reqrow)<cols:
                for c in range(len(reqrow),cols):
                    reqrow.append(self._typematch(None, self.field_type[c]))
            # add the row to value
            self._value.append(reqrow)

    value = property(get_value, set_value)

    @property
    def string_value(self):
        "Converts list of idents and WidgFields to list of strings, Textblocks to textref's"
        table = []
        row_list = []
        for row in self._value:
            row_list = []
            for idx, item in enumerate(row):
                if self.field_type[idx] == 'textblock_ref':
                    row_list.append(item.textref)
                elif self.field_type[idx] == 'boolean':
                    if item:
                        row_list.append("True")
                    else:
                        row_list.append("False")
                elif item is None:
                    row_list.append('')
                else:
                    row_list.append(str(item))
            table.append(row_list)
        return table


    def __len__(self):
        "Returns the number of rows"
        return len(self._value)

    def __bool__(self):
        "Returns True if any item in the table is boolean true, otherwise returns False"
        for row in self._value:
            for col in row:
                if col:
                    return True
        return False


class FieldArgDict(ParentFieldArg):
    "A field which takes a dictionary of items"

    def __init__(self, field_type, valdt=False, jsonset=False, senddict = False):
        """A dictionary
        """
        ParentFieldArg.__init__(self, field_type, valdt=valdt, jsonset=jsonset, senddict=senddict)
        self._value = collections.OrderedDict()

    def __getitem__(self, index):
        return self._value[index]

    def __setitem__(self, index, val):
        self._value[index] = val

    def __delitem__(self, index):
        del self._value[index]

    def get_value(self):
        return self._value

    def set_value(self, val):
        # val should be a either a dictionary or an ordered dictionary
        self._value = collections.OrderedDict()
        if not val:
            return
        if not isinstance(val, dict):
            raise ValidateError("Invalid value")
        if isinstance(val, collections.OrderedDict):
            value_list = list(val.items())
        else:
            value_list = sorted(val.items(), key=lambda t: t[0])
        # value_list is a list of tuples [(key, value), ... ]
        typed_list = [ (key, self._typematch(value, self.field_type)) for key,value in value_list ]
        self._value = collections.OrderedDict(typed_list)

    value = property(get_value, set_value)

    @property
    def string_value(self):
        "Converts ordered dict of idents and WidgFields to ordered dict of strings, Textblocks to textref's"
        value = collections.OrderedDict()
        for key, val in self._value.items():
            if self.field_type == 'textblock_ref':
                value[key] = val.textref
            elif self.field_type == 'boolean':
                if val:
                    value[key] = "True"
                else:
                    value[key] = "False"
            elif val is None:
                value[key] = ''
            else:
                value[key] = str(val)
        return value

    def __len__(self):
        return len(self._value)

    def __bool__(self):
        "Returns True if any item in the dictionarey is boolean true, otherwise returns False"
        for val in self._value.values():
            if val:
                return True
        return False


def _create_validator_list(val_list, proj_ident):
    v_list = []
    for validator in val_list:
        val_dict = collections.OrderedDict()
        v_mod = validator.__module__.split(".")[-1]
        val_dict['class'] = "%s.%s" % (v_mod, validator.__class__.__name__)
        # now write the validator fields
        if validator.message:
            val_dict["message"] = validator.message
        if validator.message_ref:
            val_dict["message_ref"] = validator.message_ref
        if validator.displaywidget:
            val_dict["displaywidget"] = validator.displaywidget.to_tuple()
        allowed_values = validator.allowed_values
        if allowed_values:
            val_dict["allowed_values"] = tag.make_list(allowed_values, proj_ident)
        val_args = {}
        if validator.val_args:
            val_args = collections.OrderedDict(sorted(validator.val_args.items(), key=lambda t: t[0]))
            val_dict["val_args"] = tag.make_dictionary(val_args, proj_ident)
        v_list.append(val_dict)
    return v_list



############################################################
#
# The Widget class - subclassed to produce widgets
#
############################################################


class Widget(tag.Part):
    """
    Inherits from tag.Part, and should be subclassed to produce widgets.
    """

    # If this widget can contain further widgets or parts, these container locations will be held within
    # tuple _container, which is a tuple of location tuples, giving the location of the container within the widget
    # that is, the location tuple is relative to the widget, not the page.

    # This container is the widget itself
    # _container=((), )

    _container = None
    # _container is set to None if no container is available in the widget

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=False

    # display_errors is a class attribute, True if this widget accepts and displays error messages, False otherwise
    display_errors = True

    # error_location is either None, or an integer or tuple, pointing to an error location,
    # if it is not None, then show_error will show the error at that location
    error_location = None

    arg_descriptions = {'widget_class':FieldArg("cssclass", "", jsonset=True)}

    # All widgets automatically have a show, show_error, widget_class and widget_style field arguments

    def __init__(self, name=None, tag_name="div", brief='', **field_args):

        if not brief:
            brief = self.__class__.__name__

        tag.Part.__init__(self, tag_name=tag_name, text='', show=True, brief=brief, hide_if_empty=False)
        self.name = name
        # create self.fields which is a dictionary of {field_arg:FieldArg,...}
        # for example - in this case field_arg is the string 'widget_class', and value is
        # a deep copy of the FieldArg object created by FieldArg("cssclass", "", jsonset=True)
        self.fields = copy.deepcopy(self.arg_descriptions)
        # ALL Widgets have a show and widget_class arguments
        if 'show' not in self.fields:
            self.fields['show'] = FieldArg("boolean", True)
        if 'widget_class' not in self.fields:
            self.fields['widget_class'] = FieldArg("cssclass", '', jsonset=True)
        if 'widget_style' not in self.fields:
            self.fields['widget_style'] = FieldArg("cssstyle", '', jsonset=True)
        # if widget has display_errors set True, then they also have a show_error and clear_error argument
        if self.display_errors:
            if 'show_error' not in self.fields:
                self.fields['show_error'] = FieldArg("text", '', jsonset=True)
            if 'clear_error' not in self.fields:
                self.fields['clear_error'] = FieldArg("boolean", False, jsonset=True )
        # Each FieldArg object has a name and value.
        # **field_args is a dictionary of {field_arg:value, ...}  - in this example it could be
        # {"widget_class": "css_class_name',...}
        # For each FieldArg object in self.fields, if its field_arg does not appear in the **field_args, then its value will be that
        # given in the class attribute 'default'. If it does appear, then its
        # value is set to that given in **field_args
        for field_arg, field in self.fields.items():
            # initially set field names to be the same as field argument
            if self.name:
                field.name = field_arg
            else:
                # If this widget has no name - it is set inside another. To avoid name clashes when inserting one widget in another,
                # all names in this widget are initially set to start with an underscore"
                field.name = '_' + field_arg
            # set field values to be the values given in **field_args (note: field_args is plural to differentiate)
            if field_arg in field_args:
                field.value = field_args[field_arg]
        # set initial self.error_message to the value given in show_error argument
        if self.display_errors and self.fields['show_error'].value:
            self.error_message = self.fields['show_error'].value
        else:
            self.error_message = ''

        # the widget show is set by the show argument
        self.show = self.fields["show"].value

        # Creates a dictionary of names against field arguments in self._names
        self._create_name_dict()

        # If the widget is in error status, this becomes True
        self.error_status = False

    @classmethod
    def module_name(cls):
        return cls.__module__.split('.')[-1]

    def _create_name_dict(self):
        "Creates a dictionary of names against field arguments"
        self._names = { field.name:field_arg for field_arg, field in self.fields.items() }


    def get_container_parts(self, index):
        """"index is the index in the self._container list
           If index out of range, return None"""
        container_part = self.container_part(index)
        if container_part is None:
            return
        return container_part.parts


    def append_to_container(self, index, value):
        """appends value into the container, in this case index is not the part location,
           it is the index in the self._container list"""
        container_part = self.container_part(index)
        if container_part is None:
            return
        if (len(container_part.parts) == 1) and (container_part.parts[0] == ''):
            # replace the empty string
            container_part.parts[0] = value
        else:
            # append the value to the container
            container_part.append(value)


    def is_container_empty(self, index):
        """"index is the index in the self._container list
           If index out of range, return None, otherwise True if empty, False if not"""
        container_part = self.container_part(index)
        if container_part is None:
            return
        if container_part.parts:
            # has contents
            if (len(container_part.parts) == 1) and (container_part.parts[0] == ""):
                return True
        else:
            return True
        return False


    @classmethod
    def can_contain(cls):
        if cls._container is None:
            return False
        return True


    def container_part(self, index):
        "Returns element, genearlly a div which is the container holding element, used internally"
        if self._container is None:
            return
        if (index < 0) or (index >= len(self._container)):
            return
        location = self._container[index]
        if location == ():
            # the container is the widget itself
            return self
        return self.get_location_value(location)


    def set_in_container(self, index, location, value):
        """Sets a value within a container, index is the container index
           and location is an integer or tuple within the container.
           for example index=0, location=(0,1) referes to location (0,1) inside container 0"""
        container_part = self.container_part(index)
        if container_part is None:
            return
        container_part.set_location_value(location, value)


    def insert_into_container(self, index, position, value):
        """Inserts a value within a container at position, index is the container index
           and position is an integer within the container.
           for example index=0, position=0 referes to the first position inside container 0"""
        container_part = self.container_part(index)
        if container_part is None:
            return
        if not isinstance(position, int):
            return
        container_part.insert(position, value)        


    def get_from_container(self, index, location):
        """gets the value from within a container, index is the container index
           and location is an integer or tuple within the container.
           for example index=0, location=(0,1) referes to location (0,1) inside container 0"""
        container_part = self.container_part(index)
        if container_part is None:
            return
        # empty container contains an empty string
        if self.is_container_empty(index) and ((not location) or (location == (0,)) or (location == [0])):
            return ''
        return container_part.get_location_value(location)


    def del_from_container(self, index, location):
        """Deletes the value from within a container, index is the container index
           and location is an integer or tuple within the container.
           for example index=0, location=(0,1) referes to location (0,1) inside container 0"""
        container_part = self.container_part(index)
        if container_part is None:
            return
        container_part.del_location_value(location)


    @classmethod
    def len_containers(cls):
        "Returns number of containers"
        if cls._container is None:
            return 0
        return len(cls._container)


    @classmethod
    def get_container_loc(cls, index):
        """Returns the location tuple of the container, in this case index is not
           the part location, it is the index in the self._container list"""
        if cls._container is None:
            return
        if (index < 0) or (index >= len(cls._container)):
            return
        return cls._container[index]


    @classmethod
    def get_container_string_loc(cls, index):
        """Returns the string location of the container, in this case index is not
           the part location, it is the index in the self._container list"""
        if cls._container is None:
            return
        if (index < 0) or (index >= len(cls._container)):
            return
        location = cls._container[index]
        if not location:
            return ''
        return '-'.join(str(i) for i in location)


    def get_parent_widget(self, page_or_section):
        """Returns the parent widget and container index of this widget"""
        section_name, parent_widget_name, parent_container = self.embedded
        if not parent_widget_name:
            return None, None
        # Find parent widget of this widget
        widgets_dict = page_or_section.widgets
        if parent_widget_name in widgets_dict:
            return widgets_dict[parent_widget_name], parent_container
        else:
            return None, None


    @classmethod
    def field_arguments_single(cls):
        """Returns a list of lists.
        The inner list consists of: [ field arg, field type, valdt, jsonset, cssclass, cssstyle]
        sorted by field argument"""
        args = []
        for arg, item in cls.arg_descriptions.items():
            if isinstance(item, FieldArg):
                args.append( [arg, item.field_type, item.valdt, item.jsonset, item.cssclass, item.cssstyle] )
        if 'show' not in cls.arg_descriptions:
            args.append( ['show', 'boolean', False, False, False, False] )
        if 'widget_class' not in cls.arg_descriptions:
            args.append( ['widget_class', 'cssclass', False, True, True, False] )
        if 'widget_style' not in cls.arg_descriptions:
            args.append( ['widget_style', 'cssstyle', False, True, False, True] )
        if cls.display_errors and ('show_error' not in cls.arg_descriptions):
            args.append( ['show_error', 'text', False, True, False, False] )
        if cls.display_errors and ('clear_error' not in cls.arg_descriptions):
            args.append( ['clear_error', 'boolean', False, True, False, False] )
        args.sort(key=lambda row: row[0])
        return args


    @classmethod
    def field_arguments_list(cls):
        """Returns a list of lists.
        The inner list consists of: [ field arg, field type, valdt, jsonset].
        sorted by field argument"""
        arg_list = []
        for arg, item in cls.arg_descriptions.items():
            if isinstance(item, FieldArgList):
                arg_list.append( [arg, item.field_type, item.valdt, item.jsonset] )
        if arg_list:
            arg_list.sort(key=lambda row: row[0])
        return arg_list


    @classmethod
    def field_arguments_table(cls):
        """Returns a list of lists.
        The inner list consists of: [ field arg, field type, valdt, jsonset].
        sorted by field argument.
        field type is a list of column types"""
        arg_table = []
        for arg, item in cls.arg_descriptions.items():
            if isinstance(item, FieldArgTable):
                arg_table.append( [arg, item.field_type, item.valdt, item.jsonset] )
        if arg_table:
            arg_table.sort(key=lambda row: row[0])
        return arg_table


    @classmethod
    def field_arguments_dictionary(cls):
        """Returns a list of lists.
        The inner list consists of: [ field arg, field type, valdt, jsonset].
        sorted by field argument"""
        arg_dict = []
        for arg, item in cls.arg_descriptions.items():
            if isinstance(item, FieldArgDict):
                arg_dict.append( [arg, item.field_type, item.valdt, item.jsonset] )
        if arg_dict:
            arg_dict.sort(key=lambda row: row[0])
        return arg_dict


    @classmethod
    def field_arguments(cls):
        """Returns a list of field args"""
        args = []
        for arg in cls.arg_descriptions:
            args.append(arg)
        if 'show' not in cls.arg_descriptions:
            args.append('show')
        if 'widget_class' not in cls.arg_descriptions:
            args.append('widget_class')
        if 'widget_style' not in cls.arg_descriptions:
            args.append('widget_style')
        if cls.display_errors and ('show_error' not in cls.arg_descriptions):
            args.append('show_error')
        if cls.display_errors and ('clear_error' not in cls.arg_descriptions):
            args.append('clear_error')
        args.sort()
        return args


    def field_arg_info(self, dataarg):
        """Returns (field name, fieldvalue, str_fieldvalue, fieldarg class string, field type, field.valdt, field.jsonset, field.cssclass, field.cssstyle)
           where fieldarg class string is one of the strings 'args', 'arg_list', 'arg_table', 'arg_dict'
           returns an empty tuple if not found"""
        if not dataarg in self.fields:
            return ()
        field = self.fields[dataarg]
        if isinstance(field, FieldArg):
            fieldarg = 'args'
        elif isinstance(field, FieldArgList):
            fieldarg = 'arg_list'
        elif isinstance(field, FieldArgTable):
            fieldarg = 'arg_table'
        elif isinstance(field, FieldArgDict):
            fieldarg = 'arg_dict'
        else:
            return ()
        return (field.name, field.value, field.string_value, fieldarg, field.field_type, field.valdt, field.jsonset, field.cssclass, field.cssstyle)


    def field_arg_val_list(self, dataarg):
        "Returns the list of validators attached to this field"
        if not dataarg in self.fields:
            return []
        field = self.fields[dataarg]
        return field.val_list

    def changed_names(self):
        "Returns a dictionary of those arguments whose name is different to the argument"
        return {a:n for n,a in self._names.items() if n != a}


    def _build(self, page, ident_list, environ, call_data, lang):
        "Called by update - the dynamic parts of the widget should be created here"
        return


    def update(self, page, ident_list, environ, call_data, lang, ident_string, placename, embedded):
        """Runs self._build, then update all sub widgets"""
        if not self.show:
            return
        if placename:
            self.placename = placename
        if not self.ident_string:
            # item is a dynamic part, newly created within a widget
            self.embedded = embedded
            self.ident_string = ident_string
        if self.name:
            # parts beneath this named widget will be embedded
            embedded_parts = (self.embedded[0], self.name, None)
        else:
            embedded_parts = self.embedded
        if self._error:
            if isinstance(self._error, TextBlock):
                self._error.update(page, ident_list, environ, call_data, lang, self.ident_string, self.placename, embedded_parts)
            return
        # build the widget
        # the class attribute is set by 'widget_class'
        if self.fields['widget_class'].value:
            self.update_attribs({'class':self.fields['widget_class'].value})
        if self.fields['widget_style'].value:
            self.update_attribs({'style':self.fields['widget_style'].value})
        # Insert this widgets id
        self.insert_id()
        # insert further parts according to each widget build
        self._build(page, ident_list, environ, call_data, lang)
        # update all parts, including those created by self._build
        try:
            for index, part in enumerate(self.parts):
                part_ident_string = self.ident_string + "-" + str(index)
                if hasattr(part, "update"):
                    part.update(page, ident_list, environ, call_data, lang, part_ident_string, self.placename, embedded_parts)
        except ValidateError as e:
            if not e.ident_list:
                e.ident_list = ident_list
            raise


    def _make_fieldvalues(self, *fieldargs, **otherparams):
        "Creates a javascript string of fieldvalues, which can be used by _build_js"
        fieldvalues = {}
        for farg in fieldargs:
            fieldvalues[farg] = self.get_field_value(farg)
        fieldvalues.update(otherparams)
        if not fieldvalues:
            return ''
        return """  SKIPOLE.widgets["{ident}"].fieldvalues={fieldvalues};
""".format(ident=self.get_id(), fieldvalues=json.dumps(fieldvalues))

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Called by make_js, and should be overwritten by widgets to set widget specific javascript
           into the page.  This javascript is added last, after all other widget updates are done, and
           this function is only called if the widget has a name."""
        return ''

    def make_js(self, page, ident_list, environ, call_data, lang):
        """Called by page.update - after parts update is called.
           Makes javascript for this widget, calls make_js for sub parts/widgets, then if widget
           has a name, calls self._build_js to build further javascript for this widget"""
        if self._error:
            return
        if not self.show:
            return
        # First, call make_js of sub widgets
        tag.Part.make_js(self, page, ident_list, environ, call_data, lang)

        if not self.name:
            return

        # js_validators is a class attribute, True is javascript validation is enabled
        if self.js_validators and not skiboot.get_debug():
            # Add the javascript validators for this widget,
            # adds a line for each widgfield that has validators
            # SKIPOLE.validators[widgfield] = [[...], [...], ... ]
            # where each of the lists correspond to a validator for this widgfield
            # [val_mod, val_class, message, [allowed values], {args}]
            for field_arg, field in self.fields.items():
                all_validators_list = []
                val_list = field.val_list
                if not val_list:
                    # no validators to add on this field
                    continue
                widgfield = self.name + ':' + field.name
                if self.placename:
                    widgfield = self.placename + '-' + widgfield
                for val in val_list:
                    one_validator_list = []
                    one_validator_list.append(val.__class__.__module__.split(".")[-1])
                    one_validator_list.append(val.__class__.__name__)
                    message = val.get_message(lang, page.proj_ident)
                    if not message:
                        message = ""
                    one_validator_list.append(message)
                    one_validator_list.append(val.allowed_values)
                    one_validator_list.append(val.val_args)
                    one_validator_list.append(str(val.displaywidget))
                    one_validator_list
                    all_validators_list.append(one_validator_list)
                if all_validators_list:
                    v = "  SKIPOLE.validators[\"%s\"] = " % (widgfield,)
                    v += json.dumps(all_validators_list)
                    page.add_javascript(v+';\n')

        # information about the widget, its id, module, class and arguments
        j = Template("""  SKIPOLE.widgets["$widgident"] = new SKIPOLE["$widg_mod"]["$widg_class"]("$widgident", $error_message, $widg_fields);
""")
        page.add_javascript(j.substitute(widgident = self.get_id(),
                                         widg_mod=self.__class__.__module__.split(".")[-1],
                                         widg_class=self.__class__.__name__,
                                         error_message = json.dumps(self.error_message),
                                         widg_fields=str(self.changed_names())))

        # and add widget specific json
        contents = self._build_js(page, ident_list, environ, call_data, lang)
        if contents:
            page.add_javascript(contents)


    def set_placename(self, section_name, placename):
        "Widgets in sections with displayname validators need displaynames to change"
        self.placename = placename
        for field in self.fields.values():
            if field.val_list:
                # the field has a list of validators
                val_list = field.val_list
                for val in val_list:
                    # If the displaywidget for a validator is equal to the section_name
                    # replace it with the placename
                    if val.displaywidget.s == section_name:
                        val.displaywidget = val.displaywidget._replace(s=placename)

    def mark_field_in_error(self, errorfieldname):
        "Marks a given field as an errored field, for example, by setting it red, child classes may optionally implement this"
        pass

    def _error_build(self, message):
        """Overwritten by child widgets if required, note this is run before the _build method
           and should set any appropriate flags that _build will then use to build the widget"""
        if self.error_location is None:
            # no error location set, so replace entire widget by the message
            if message:
                self._error = message


    def show_error(self, message=''):
        """Shows error message"""

        if not self.show:
            return
        if not self.display_errors:
            return
        self.error_status = True
        # if this is a named widget update attribute with data-status="error"
        if self.name:
            self.update_attribs({'data-status':'error'})
        if not message:
            message = self.error_message
        if self.error_location is not None:
            error_part = self.get_location_value(self.error_location)
            if hasattr(error_part, 'show_error'):
                error_part.show_error(message)
            elif message:
                self.set_location_value(self.error_location, message)
        # and call _error_build
        self._error_build(message)
    
    def field_list(self):
        "Returns a list of fields, ordered alphabetically by field argument"
        return [ self.fields[key] for key in sorted(self.fields.keys())]


    def get_field(self, name):
        "Given a name, get the associated field"
        if not name:
            return
        if name in self._names:
            field_arg = self._names[name]
            if field_arg in self.fields:
                field = self.fields[field_arg]
                if field.name == name:
                    return field
        # failed to find field, refresh self._names and try again
        self._create_name_dict()
        if name in self._names:
            field_arg = self._names[name]
            if field_arg in self.fields:
                field = self.fields[field_arg]
                if field.name == name:
                    return field

    def get_field_arg(self, name):
        "Given a name, get the associated field_arg, returns None if name not found"
        if name in self._names:
            return self._names[name]


    def get_field_valdt(self, name):
        "Given a name, get field.valdt status, True if validators can be added, False if not"
        field = self.get_field(name)
        if field is None:
            return False
        return field.valdt


    def is_fieldname_in_widget(self, name):
        field = self.get_field(name)
        if field is None:
            return False
        return True


    def is_senddict(self, name):
        field = self.get_field(name)
        if field is None:
            raise ValidateError(message = "Widget %s field name %s not recognised" % (self.name, name))
        return field.senddict
        

    def add_validator(self, field_name, val):
        "adds a validator to the field with the given name"
        f = self.get_field(field_name)
        if f is not None:
            f.add_validator(val)

    def validate(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        "Calls the field validate method, returns item, error_list"
        f = self.get_field(widgfield.f)
        if f is None:
            raise ValidateError(message = "Field name not recognised")
        return f.validate(widgfield, item, environ, lang, form_data, call_data, caller_page_ident)

    def get_value(self, name):
        """If this widget contains fields, then this should return a field
        content, given its name.  If name not found, return None
        """
        field = self.get_field(name)
        if field is not None:
            return field.value

    def set_value(self, name, value):
        """If this widget contains fields, then this should set a field
           content, given its name."""
        if not name:
            return
        if '-' in name:
            self.set_single_multivalue(name, value)
            return
        if (name == 'show_error') or (('show_error' in self.fields) and (name == self.fields['show_error'].name)):
            # the show_error field is special as it accepts both a given field name
            # and the name 'show_error'
            if value:
                self.show_error(message=value)
            else:
                self.show_error()
            return
        field = self.get_field(name)
        if field is None:
            return
        field.value = value
        if name == self.fields["show"].name:
            self.show = bool(value)
        if name == self.fields['widget_class'].name:
            if value:
                self.update_attribs({'class':value})
            else:
                self.del_one_attrib('class')
        if name == self.fields['widget_style'].name:
            if value:
                self.update_attribs({'style':value})
            else:
                self.del_one_attrib('style')

        

    def set_single_multivalue(self, name, value):
        """If a field name is given with a -, then after the dash must be an index
           This sets the single value with the appropriate index"""
        if '-' not in name: return
        fname, findex = name.split('-')
        field = self.get_field(fname)
        if field is None:
            return
        if not isinstance(field, FieldArgDict):
            return
        field[findex] = value

    def _set_data(self, *loc):
        """Clears attributes of the part at location loc, then sets n data- atts"""
        # get the number of fields with arguments data1, data2,..
        n = 0   # increments for each data field
        m = 0   # records max number
        for field_arg in self.fields:
            if len(field_arg) < 5: continue
            if field_arg.startswith('data'):
                endnumber = field_arg[4:]
                try:
                    num = int(endnumber)
                except Exception:
                    continue
                if not num: continue
                if num > m:
                    m = num
                n += 1
        if not n:
            # no field_arg's with name data1,...
            return
        if n != m:
            raise ServerError("field data values are mismatched")
        if loc:
            part = self.get_location_value(loc)
        else:
            part = self
        part.del_attribs()
        for idx in range(n):
            # field_arg is data1, data2 etc..
            field_arg = 'data' + str(idx+1)
            if field_arg not in self.fields:
                raise ServerError("Field argument %s does not exist in this widget." % (field_arg,))
            if self.get_field_value(field_arg):
                key = 'data-' + self.get_formname(field_arg)
                part.update_attribs({key:str(self.get_field_value(field_arg))})

    def set_field_value(self, field_arg, value):
        """If this widget contains fields, then this should set a field
           content, given its field_arg"""
        if not field_arg:
            return
        if field_arg not in self.fields:
            return
        field = self.fields[field_arg]
        if field is None:
            return
        field.value = value
        if field_arg == "show":
            self.show = field.value
        if field_arg == 'widget_class':
            if value:
                self.update_attribs({'class':value})
            else:
                self.del_one_attrib('class')
        if field_arg == 'widget_style':
            if value:
                self.update_attribs({'style':value})
            else:
                self.del_one_attrib('style')

        if field_arg == 'show_error':
            self.error_message = field.value

    def get_field_value(self, field_arg):
        """If this widget contains fields, then this should get a field
           content, given its field_arg"""
        if not field_arg:
             raise ValidateError(message="A valid field argument is required")
        if field_arg not in self.fields:
            raise ValidateError(message="Field argument %s not recognised in %s" % (field_arg, self.__class__.__name__))
        return self.fields[field_arg].value

    def set_name(self, field_arg, name):
        """Sets a field name, raises ValidateError if the name already exists in the widget"""
        # check name does not already exist
        if not name:
            return
        if name.lower() == "ident":
            raise ValidateError(message="Field name ident is used internally and is not allowed")
        if name == 'tag_name':
            raise ValidateError(message="Field name tag_name is used internally and is not allowed")
        if name == 'error_message':
            raise ValidateError(message="Field name error_message is used internally and is not allowed")
        if _AN.search(name):
            raise ValidateError(message="Invalid name - letters, numbers, underscore only.")
        if field_arg not in self.fields:
            return
        if self.fields[field_arg].name == name:
            # no change
            return
        # check name does not already exist in the widget
        if self.is_fieldname_in_widget(name):
            # a field has been found
            raise ValidateError(message="This field name already exists in the widget")
        # set the name
        self.fields[field_arg].name = name
        # refresh self._names
        self._create_name_dict()

    def get_name(self, field_arg):
        """Gets a field name"""
        if not field_arg:
             raise ValidateError(message="A valid field argument is required")
        if field_arg not in self.fields:
            raise ValidateError(message="Field argument not recognised")
        return self.fields[field_arg].name

    def get_formname(self, field_arg):
        """Returns the string 'sectionname-widgetname:fieldname', used to set submission field names"""
        fieldname = self.get_name(field_arg)
        if not fieldname:
            raise ValidateError(message="Invalid field name")
        if self.name:
            fieldname = self.name + ':' + fieldname
        elif self.embedded[1]:
            fieldname = self.embedded[1] + ':' + fieldname
        else:
            raise ValidateError(message="Widget \"%s\", has no name." % (self.__class__.__name__,))
        if self.placename:
            fieldname = self.placename + '-' + fieldname
        return fieldname


    def add_hiddens(self, form, page=None):
        "Used to add ident and four hidden fields to a form, requires this widget to have these hidden fields"
        # all submissions always have an 'ident' hidden field to provide the ident of the calling page
        if page is not None:
            form.append(tag.ClosedPart(tag_name="input",
                                   attribs ={"name":'ident',
                                             "value":page.ident_data_string,
                                             "type":"hidden"}))
        # hidden field on the form
        if self.get_field_value('hidden_field1'):
            form.append(tag.ClosedPart(tag_name="input",
                                       attribs ={"name":self.get_formname('hidden_field1'),
                                                 "value":self.get_field_value('hidden_field1'),
                                                 "type":"hidden"}))

        # Second hidden field on the form
        if self.get_field_value('hidden_field2'):
            form.append(tag.ClosedPart(tag_name="input",
                                       attribs ={"name":self.get_formname('hidden_field2'),
                                                 "value":self.get_field_value('hidden_field2'),
                                                 "type":"hidden"}))

        # third hidden field on the form
        if self.get_field_value('hidden_field3'):
            form.append(tag.ClosedPart(tag_name="input",
                                       attribs ={"name":self.get_formname('hidden_field3'),
                                                 "value":self.get_field_value('hidden_field3'),
                                                 "type":"hidden"}))
        # fourth hidden field on the form
        if self.get_field_value('hidden_field4'):
            form.append(tag.ClosedPart(tag_name="input",
                                       attribs ={"name":self.get_formname('hidden_field4'),
                                                 "value":self.get_field_value('hidden_field4'),
                                                 "type":"hidden"}))


    def widget_hide(self, hide):
        "Hides widget if not error"
        if (not self.error_status) and hide:
            self.set_hide()
        else:
            self.set_block()


    def __repr__(self):
        if self.name:
            return self.__class__.__name__ + '(' + self.name + ')'
        else:
            return self.__class__.__name__


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<div>  <!-- with widget id and class widget_class -->
</div>"""


    def __str__(self):
        """Returns a text string to illustrate the widget"""
        return self.description()


    def outline(self, proj_ident):
        part_dict = collections.OrderedDict()
        w_mod = self.__module__.split(".")[-1]
        part_dict['class'] = "%s.%s" % (w_mod, self.__class__.__name__)
        if self.name:
            part_dict["name"] = self.name
        if self.brief:
            part_dict["brief"] = self.brief
        fields_dict = {f_arg: f.value for f_arg, f in self.fields.items()}
        if fields_dict:
            ordered_fields_dict = collections.OrderedDict(sorted(fields_dict.items(), key=lambda t: t[0]))
            part_dict["fields"] = tag.make_dictionary(ordered_fields_dict, proj_ident)
        # set widget containers
        if self.can_contain():
            for cont in range(self.len_containers()):
                container_name = "container_%s" % (cont,)
                # get list of parts in the container
                parts = self.get_container_parts(cont)
                item_list = []
                for item in parts:
                    if hasattr(item, 'outline'):
                        item_list.append(item.outline(proj_ident))
                    else:
                        # must be a text string
                        item_list.append(['Text', str(item)])
                part_dict[container_name] = item_list
        # set widget field names
        if fields_dict:
            # check if any field name is not equal to f_arg
            fields_names = {f_arg:f.name for f_arg, f in self.fields.items() if f_arg != f.name}
            if fields_names:
                ordered_fields_names = collections.OrderedDict(sorted(fields_names.items(), key=lambda t: t[0]))
                part_dict["set_names"] = tag.make_dictionary(ordered_fields_names, proj_ident)
            # set widget validators
            field_validators = {f.name:f.val_list for f in self.fields.values() if f.val_list}
            if field_validators:
                val_dict = {}
                for name, val_list in field_validators.items():
                    val_dict[name] = _create_validator_list(val_list, proj_ident)
                part_dict["validators"] = collections.OrderedDict(sorted(val_dict.items(), key=lambda t: t[0]))
        return ['Widget', part_dict]





class ClosedWidget(tag.ClosedPart):
    """
    Inherits from tag.ClosedPart, and should be subclassed to produce widgets.
    """

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=False

    # display_errors is a class attribute, True if this widget accepts and displays error messages, False otherwise
    display_errors = True

    # always empty list and unused
    _container=[]

    # always None and unused
    error_location = None

    arg_descriptions = {'widget_class':FieldArg("cssclass", "", jsonset=True)}


    def __init__(self, name=None, tag_name="link", brief='', **field_args):
        if not brief:
            brief = self.__class__.__name__
        tag.ClosedPart.__init__(self, tag_name=tag_name, show=True, brief=brief)
        self.name = name
        # Get fields from arg_descriptions
        self.fields = copy.deepcopy(self.arg_descriptions)
        # ALL Widgets have a show, widget_class and widget_style arguments
        if 'show' not in self.fields:
            self.fields['show'] = FieldArg("boolean", True)
        if 'widget_class' not in self.fields:
            self.fields['widget_class'] = FieldArg("cssclass", "", jsonset=True)
        if 'widget_style' not in self.fields:
            self.fields['widget_style'] = FieldArg("cssstyle", '', jsonset=True)
        # if widget has display_errors set True, then they also have a show_error and clear_error argument
        if self.display_errors:
            if 'show_error' not in self.fields:
                self.fields['show_error'] = FieldArg("text", '', jsonset=True)
            if 'clear_error' not in self.fields:
                self.fields['clear_error'] = FieldArg("boolean", False, jsonset=True )
        for field_arg, field in self.fields.items():
            # initially set field names to be the same as field argument
            if self.name:
                field.name = field_arg
            else:
                # If this widget has no name - it is set inside another. To avoid name clashes when inserting one widget in another,
                # all names in this widget are initially set to start with an underscore"
                field.name = '_' + field_arg
            # set field values to be the values given in **field_args
            if field_arg in field_args:
                field.value = field_args[field_arg]

        # Creates a dictionary of names against field arguments in self._names
        self._create_name_dict()

        # Unlike a Widget, does not use error_location or _container
        # the widget show is set by the show argument
        self.show = self.fields["show"].value

        # set initial self.error_message to the value given in show_error argument
        if self.display_errors and self.fields['show_error'].value:
            self.error_message = self.fields['show_error'].value
        else:
            self.error_message = ''

        # If the widget is in error status, this becomes True
        self.error_status = False

    @classmethod
    def module_name(cls):
        return cls.__module__.split('.')[-1]

    @classmethod
    def can_contain(cls):
        # Always False, only provided for consistancy with Widget
        return False

    @classmethod
    def len_containers(cls):
        return 0

    def get_parent_widget(self, page_or_section):
        """Returns the parent widget and container  index of this widget"""
        section_name, parent_widget_name, parent_container = self.embedded
        if not parent_widget_name:
            return None, None
        # Find parent widget of this widget
        widgets_dict = page_or_section.widgets
        if parent_widget_name in widgets_dict:
            return widgets_dict[parent_widget_name], parent_container
        else:
            return None, None

    def _create_name_dict(self):
        "Creates a dictionary of names against field arguments"
        self._names = { field.name:field_arg for field_arg, field in self.fields.items() }


    @classmethod
    def field_arguments_single(cls):
        """Returns a list of lists.
        The inner list consists of: [ field arg, field type, valdt, jsonset, cssclass, cssstyle]
        sorted by field argument"""
        args = []
        for arg, item in cls.arg_descriptions.items():
            if isinstance(item, FieldArg):
                args.append( [arg, item.field_type, item.valdt, item.jsonset, item.cssclass, item.cssstyle] )
        if 'show' not in cls.arg_descriptions:
            args.append( ['show', 'boolean', False, False, False, False] )
        if 'widget_class' not in cls.arg_descriptions:
            args.append( ['widget_class', 'cssclass', False, True, True, False] )
        if 'widget_style' not in cls.arg_descriptions:
            args.append( ['widget_style', 'cssstyle', False, True, False, True] )
        if cls.display_errors and ('show_error' not in cls.arg_descriptions):
            args.append( ['show_error', 'text', False, True, False, False] )
        if cls.display_errors and ('clear_error' not in cls.arg_descriptions):
            args.append( ['clear_error', 'boolean', False, True, False, False] )
        args.sort(key=lambda row: row[0])
        return args


    @classmethod
    def field_arguments_list(cls):
        """Returns a list of lists.
        The inner list consists of: [ field arg, field type, valdt, jsonset].
        sorted by field argument"""
        arg_list = []
        for arg, item in cls.arg_descriptions.items():
            if isinstance(item, FieldArgList):
                arg_list.append( [arg, item.field_type, item.valdt, item.jsonset] )
        if arg_list:
            arg_list.sort(key=lambda row: row[0])
        return arg_list


    @classmethod
    def field_arguments_table(cls):
        """Returns a list of lists.
        The inner list consists of: [ field arg, field type, valdt, jsonset].
        sorted by field argument.
        field type is a list of column types"""
        arg_table = []
        for arg, item in cls.arg_descriptions.items():
            if isinstance(item, FieldArgTable):
                arg_table.append( [arg, item.field_type, item.valdt, item.jsonset] )
        if arg_table:
            arg_table.sort(key=lambda row: row[0])
        return arg_table


    @classmethod
    def field_arguments_dictionary(cls):
        """Returns a list of lists.
        The inner list consists of: [ field arg, field type, valdt, jsonset].
        sorted by field argument"""
        arg_dict = []
        for arg, item in cls.arg_descriptions.items():
            if isinstance(item, FieldArgDict):
                arg_dict.append( [arg, item.field_type, item.valdt, item.jsonset] )
        if arg_dict:
            arg_dict.sort(key=lambda row: row[0])
        return arg_dict


    @classmethod
    def field_arguments(cls):
        """Returns a list of field args"""
        args = []
        for arg in cls.arg_descriptions:
            args.append(arg)
        if 'show' not in cls.arg_descriptions:
            args.append('show')
        if 'widget_class' not in cls.arg_descriptions:
            args.append('widget_class')
        if 'widget_style' not in cls.arg_descriptions:
            args.append('widget_style')
        if cls.display_errors and ('show_error' not in cls.arg_descriptions):
            args.append('show_error')
        if cls.display_errors and ('clear_error' not in cls.arg_descriptions):
            args.append('clear_error')
        args.sort()
        return args


    def field_arg_info(self, dataarg):
        """Returns (field name, fieldvalue, str_fieldvalue, fieldarg class string, field type, field.valdt, field.jsonset, field.cssclass, field.csstyle)
           where fieldarg class string is one of the strings 'args', 'arg_list', 'arg_table', 'arg_dict'
           returns an empty tuple if not found"""
        if not dataarg in self.fields:
            return ()
        field = self.fields[dataarg]
        if isinstance(field, FieldArg):
            fieldarg = 'args'
        elif isinstance(field, FieldArgList):
            fieldarg = 'arg_list'
        elif isinstance(field, FieldArgTable):
            fieldarg = 'arg_table'
        elif isinstance(field, FieldArgDict):
            fieldarg = 'arg_dict'
        else:
            return ()
        return (field.name, field.value, field.string_value, fieldarg, field.field_type, field.valdt, field.jsonset, field.cssclass, field.cssstyle)

    def field_arg_val_list(self, dataarg):
        "Returns the list of validators attached to this field"
        if not dataarg in self.fields:
            return []
        field = self.fields[dataarg]
        return field.val_list

    def changed_names(self):
        "Returns a dictionary of arg:name for those arguments whose name is different to the argument"
        return {a:n for n,a in self._names.items() if n != a}

    def _build(self, page, ident_list, environ, call_data, lang):
        "Called by update - the dynamic parts of the widget should be created here"
        return

    def update(self, page, ident_list, environ, call_data, lang, ident_string, placename, embedded):
        """builds the widget"""
        if not self.show:
            return
        if placename:
            self.placename = placename
        if not self.ident_string:
            # item is a dynamic part, newly created within a widget
            self.embedded = embedded
            self.ident_string = ident_string
        if self.name:
            # parts beneath this named widget will be embedded
            embedded_parts = (self.embedded[0], self.name, None)
        else:
            embedded_parts = self.embedded
        if self._error:
            if isinstance(self._error, TextBlock):
                self._error.update(page, ident_list, environ, call_data, lang, self.ident_string, self.placename, embedded_parts)
            return
        # build the widget
        # the class attribute is set by 'widget_class'
        if self.fields['widget_class'].value:
            self.update_attribs({'class':self.fields['widget_class'].value})
        if self.fields['widget_style'].value:
            self.update_attribs({'style':self.fields['widget_style'].value})
        # Insert this widgets id
        self.insert_id()
        self._build(page, ident_list, environ, call_data, lang)

    def _make_fieldvalues(self, *fieldargs, **otherparams):
        "Creates a javascript string of fieldvalues, which can be used by _build_js"
        fieldvalues = {}
        for farg in fieldargs:
            fieldvalues[farg] = self.get_field_value(farg)
        fieldvalues.update(otherparams)
        if not fieldvalues:
            return ''
        return """  SKIPOLE.widgets["{ident}"].fieldvalues={fieldvalues};""".format(ident=self.get_id(), fieldvalues=json.dumps(fieldvalues))

    def _build_js(self, page, ident_list, environ, call_data, lang):
        return ''

    def make_js(self, page, ident_list, environ, call_data, lang):
        """Called by page.update - after parts update is called.
           Makes javascript for this widget, calls self._build_js to build further javascript for this widget"""
        if self._error:
            return
        if not self.show:
            return
        if not self.name:
            return

        # js_validators is a class attribute, True is javascript validation is enabled
        if self.js_validators:
            # Add the javascript validators for this widget,
            # adds a line for each widgfield that has validators
            # SKIPOLE.validators[widgfield] = [[...], [...], ... ]
            # where each of the lists correspond to a validator for this widgfield
            # [val_mod, val_class, message, [allowed values], {args}]
            for field_arg, field in self.fields.items():
                all_validators_list = []
                val_list = field.val_list
                if not val_list:
                    # no validators to add on this field
                    continue
                widgfield = self.name + ':' + field.name
                if self.placename:
                    widgfield = self.placename + '-' + widgfield
                for val in val_list:
                    one_validator_list = []
                    one_validator_list.append(val.__class__.__module__.split(".")[-1])
                    one_validator_list.append(val.__class__.__name__)
                    message = val.message
                    if (not message) and val.message_ref:
                        project = skiboot.getproject(page.proj_ident)
                        if project is not None:
                            message = project.textblocks.get_text(val.message_ref, lang)
                    if not message:
                        message = ""
                    one_validator_list.append(message)
                    one_validator_list.append(val.allowed_values)
                    one_validator_list.append(val.val_args)
                    one_validator_list.append(str(val.displaywidget))
                    all_validators_list.append(one_validator_list)
                if all_validators_list:
                    v = "  SKIPOLE.validators[\"%s\"] = " % (widgfield,)
                    v += json.dumps(all_validators_list)
                    page.add_javascript(v+';\n')

        # information about the widget, its id, module, class and arguments
        j = Template("""  SKIPOLE.widgets["$widgident"] = new SKIPOLE["$widg_mod"]["$widg_class"]("$widgident", $error_message, $widg_fields);
""")
        page.add_javascript(j.substitute(widgident = self.get_id(),
                                         widg_mod=self.__class__.__module__.split(".")[-1],
                                         widg_class=self.__class__.__name__,
                                         error_message = json.dumps(self.error_message),
                                         widg_fields=str(self.changed_names())))

        # and add widget specific json
        contents = self._build_js(page, ident_list, environ, call_data, lang)
        if contents:
            page.add_javascript(contents)


    def set_placename(self, section_name, placename):
        "Widgets in sections with displayname validators need displaynames to change"
        self.placename = placename
        for field in self.fields.values():
            if field.val_list:
                # the field has a list of validators
                val_list = field.val_list
                for val in val_list:
                    # If the displaywidget for a validator is equal to the section_name
                    # replace it with the placename
                    if val.displaywidget.s == section_name:
                        val.displaywidget = val.displaywidget._replace(s=placename)


    def mark_field_in_error(self, errorfieldname):
        "Marks a given field as an errored field, for example, by setting it red, child classes may optionally implement this"
        pass

    def _error_build(self, message):
        "Overwritten by child widgets if required"
        if message:
            self._error = message
 
    def show_error(self, message=''):
        """Shows error message"""
        if not self.show:
            return
        if not self.display_errors:
            return
        self.error_status = True
        # if this is a named widget update attribute with data-status="error"
        if self.name:
            self.update_attribs({'data-status':'error'})
        if not message:
            message = self.error_message
        self._error_build(message)

    def field_list(self):
        "Returns a list of fields, ordered alphabetically by field argument"
        return [ self.fields[key] for key in sorted(self.fields.keys())]

    def get_field(self, name):
        "Given a name, get the associated field"
        if not name:
            return
        if name in self._names:
            field_arg = self._names[name]
            if field_arg in self.fields:
                field = self.fields[field_arg]
                if field.name == name:
                    return field
        # failed to find field, refresh self._names and try again
        self._create_name_dict()
        if name in self._names:
            field_arg = self._names[name]
            if field_arg in self.fields:
                field = self.fields[field_arg]
                if field.name == name:
                    return field

    def get_field_arg(self, name):
        "Given a name, get the associated field_arg, returns None if name not found"
        if name in self._names:
            return self._names[name]

    def get_field_valdt(self, name):
        "Given a name, get field.valdt status, True if validators can be added, False if not"
        field = self.get_field(name)
        if field is None:
            return False
        return field.valdt

    def is_fieldname_in_widget(self, name):
        field = self.get_field(name)
        if field is None:
            return False
        return True

    def is_senddict(self, name):
        field = self.get_field(name)
        if field is None:
            raise ValidateError(message = "Widget %s field name %s not recognised" % (self.name, name))
        return field.senddict

    def add_validator(self, field_name, val):
        "adds a validator to the field with the given name"
        f = self.get_field(field_name)
        if f is not None:
            f.add_validator(val)

    def validate(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        "Calls the field validate method, returns item, error_list"
        f = self.get_field(widgfield.f)
        if f is None:
            raise ValidateError(message = "Field name not recognised")
        return f.validate(widgfield, item, environ, lang, form_data, call_data, caller_page_ident)

    def get_value(self, name):
        """If this widget contains fields, then this should return a field
        content, given its name.  If name not found, return None
        """
        field = self.get_field(name)
        if field is not None:
            return field.value

    def set_value(self, name, value):
        """If this widget contains fields, then this should set a field
           content, given its name"""
        if not name:
            return
        if '-' in name:
            self.set_single_multivalue(name, value)
            return
        if (name == 'show_error') or (('show_error' in self.fields) and (name == self.fields['show_error'].name)):
            # the show_error field is special as it accepts both a given field name
            # and the name 'show_error'
            if value:
                self.show_error(message=value)
            else:
                self.show_error()
            return
        field = self.get_field(name)
        if field is None:
            return
        field.value = value
        if name == self.fields["show"].name:
            self.show = bool(value)
        if name == self.fields['widget_class'].name:
            if value:
                self.update_attribs({'class':value})
            else:
                self.del_one_attrib('class')
        if name == self.fields['widget_style'].name:
            if value:
                self.update_attribs({'style':value})
            else:
                self.del_one_attrib('style')


    def set_single_multivalue(self, name, value):
        """If a field name is given with a -, then after the dash must be an index
           This sets the single value with the appropriate index"""
        if '-' not in name: return
        fname, findex = name.split('-')
        field = self.get_field(fname)
        if field is None:
            return
        if not isinstance(field, FieldArgDict):
            return
        field[findex] = value

    def _set_data(self):
        """Clears attributes, then sets n data- atts"""
        # get the number of fields with arguments data1, data2,..
        n = 0   # increments for each data field
        m = 0   # records max number
        for field_arg in self.fields:
            if len(field_arg) < 5: continue
            if field_arg.startswith('data'):
                endnumber = field_arg[4:]
                try:
                    num = int(endnumber)
                except Exception:
                    continue
                if not num: continue
                if num > m:
                    m = num
                n += 1
        if not n:
            # no field_arg's with name data1,...
            return
        if n != m:
            raise ServerError("field data values are mismatched")
        self.del_attribs()
        for idx in range(n):
            # field_arg is data1, data2 etc..
            field_arg = 'data' + str(idx+1)
            if field_arg not in self.fields:
                raise ServerError("Field argument %s does not exist in this widget." % (field_arg,))
            if self.get_field_value(field_arg):
                key = 'data-' + self.get_formname(field_arg)
                self.update_attribs({key:str(self.get_field_value(field_arg))})

    def set_field_value(self, field_arg, value):
        """If this widget contains fields, then this should set a field
           content, given its field_arg"""
        if not field_arg:
            return
        if field_arg not in self.fields:
            return
        field = self.fields[field_arg]
        if field is None:
            return
        field.value = value
        if field_arg == "show":
            self.show = field.value
        if field_arg == 'widget_class':
            if value:
                self.update_attribs({'class':value})
            else:
                self.del_one_attrib('class')
        if field_arg == 'widget_style':
            if value:
                self.update_attribs({'style':value})
            else:
                self.del_one_attrib('style')
        if field_arg == 'show_error':
            self.error_message = field.value

    def get_field_value(self, field_arg):
        """If this widget contains fields, then this should get a field
           content, given its field_arg"""
        if not field_arg:
             raise ValidateError(message="A valid field argument is required")
        if field_arg not in self.fields:
            raise ValidateError(message="Field argument not recognised")
        return self.fields[field_arg].value

    def set_name(self, field_arg, name):
        """Sets a field name, raises ValidateError if the name already exists in the widget"""
        # check name does not already exist
        if not name:
            return
        if name.lower() == "ident":
            raise ValidateError(message="Field name ident is used internally and is not allowed")
        if name == 'tag_name':
            raise ValidateError(message="Field name tag_name is used internally and is not allowed")
        if name == 'error_message':
            raise ValidateError(message="Field name error_message is used internally and is not allowed")
        if _AN.search(name):
            raise ValidateError(message="Invalid name - letters, numbers, underscore only.")
        if field_arg not in self.fields:
            return
        if self.fields[field_arg].name == name:
            # no change
            return
        # check name does not already exist in the widget
        if self.is_fieldname_in_widget(name):
            # a field has been found
            raise ValidateError(message="This field name already exists in the widget")
        # set the name
        self.fields[field_arg].name = name
        # refresh self._names
        self._create_name_dict()

    def get_name(self, field_arg):
        """Gets a field name"""
        if not field_arg:
             raise ValidateError(message="A valid field argument is required")
        if field_arg not in self.fields:
            raise ValidateError(message="Field argument not recognised")
        return self.fields[field_arg].name

    def get_formname(self, field_arg):
        """Returns the string 'sectionname-widgetname:fieldname', used to set submission field names"""
        fieldname = self.get_name(field_arg)
        if not fieldname:
            raise ValidateError(message="Invalid field name")
        if self.name:
            fieldname = self.name + ':' + fieldname
        elif self.embedded[1]:
            fieldname = self.embedded[1] + ':' + fieldname
        else:
            raise ValidateError(message="Widget \"%s\", has no name." % (self.__class__.__name__,))
        if self.placename:
            fieldname = self.placename + '-' + fieldname
        return fieldname

    def widget_hide(self, hide):
        "Hides widget if not error"
        if (not self.error_status) and hide:
            self.set_hide()
        else:
            self.set_block()

    def __repr__(self):
        if self.name:
            return self.__class__.__name__ + '(' + self.name + ')'
        else:
            return self.__class__.__name__

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<path />  <!-- with widget id and class widget_class -->
"""

    def __str__(self):
        """Returns a text string to illustrate the widget"""
        return self.description()


    def outline(self, proj_ident):
        part_dict = collections.OrderedDict()
        w_mod = self.__module__.split(".")[-1]
        part_dict['class'] = "%s.%s" % (w_mod, self.__class__.__name__)
        if self.name:
            part_dict["name"] = self.name
        if self.brief:
            part_dict["brief"] = self.brief
        fields_dict = {f_arg: f.value for f_arg, f in self.fields.items()}
        if fields_dict:
            ordered_fields_dict = collections.OrderedDict(sorted(fields_dict.items(), key=lambda t: t[0]))
            part_dict["fields"] = tag.make_dictionary(ordered_fields_dict, proj_ident)
            # set widget field names, check if any field name is not equal to f_arg
            fields_names = {f_arg:f.name for f_arg, f in self.fields.items() if f_arg != f.name}
            if fields_names:
                ordered_fields_names = collections.OrderedDict(sorted(fields_names.items(), key=lambda t: t[0]))
                part_dict["set_names"] = tag.make_dictionary(ordered_fields_names, proj_ident)
            # set widget validators
            field_validators = {f.name:f.val_list for f in self.fields.values() if f.val_list}
            if field_validators:
                val_dict = {}
                for name, val_list in field_validators.items():
                    val_dict[name] = _create_validator_list(val_list, proj_ident)
                part_dict["validators"] = collections.OrderedDict(sorted(val_dict.items(), key=lambda t: t[0]))
        return ['ClosedWidget', part_dict]







"""Contains widgets for lists"""


from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict



class UList1(Widget):
    """An unordered list widget."""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                       'contents': FieldArgList('text', jsonset=True),
                       'even_class':FieldArg("cssclass", ""),
                       'odd_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        contents: List of text strings to be shown, gives the number of list items
        even_class: class of even li elements, if empty string, then no class will be applied
        odd_class: class of odd li elements, if empty string, then no class will be applied
        """
        Widget.__init__(self, name=name, tag_name="ul", brief=brief, **field_args)
        self._even = ''
        self._odd = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the list"
        contents = self.get_field_value("contents")
        # set even li class
        if self.get_field_value('even_class'):
            self._even = self.get_field_value('even_class')
        # set odd li class
        if self.get_field_value('odd_class'):
            self._odd = self.get_field_value('odd_class')

        for index, item in enumerate(contents):
            if self._even and (index % 2):
                self[index] = tag.Part(tag_name="li", attribs={"class":self._even}, text=item)
            elif self._odd and not (index % 2):
                self[index] = tag.Part(tag_name='li', attribs={"class":self._odd}, text=item)
            else:
                self[index] = tag.Part(tag_name='li', text=item)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets list items"""
        return self._make_fieldvalues(even_class = self._even, odd_class = self._odd)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<ul>  <!-- with widget id and class widget_class -->
  <li> <!-- with style from odd_class or even_class -->
      <!-- with text from each contents element -->
  </li>
  <!-- with appropriate number of list elements -->
</ul>
"""








"""Contains widgets for dropdown forms"""


from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict



class DropDown1(Widget):
    """A div holding a drop down input field. Without a form or submit button, typically included within a form"""

    error_location = (0,0,0)

    arg_descriptions = {
                       'selectvalue':FieldArg("text", '', valdt=True),
                       'left_label':FieldArg("text", 'Choose:'),
                       'left_class':FieldArg("cssclass", ''),
                       'left_style':FieldArg("cssstyle", ''),
                       'error_class':FieldArg("cssclass", ""),
                       'select_class':FieldArg("cssclass", ""),
                       'select_style':FieldArg("cssstyle", ""),
                       'div_class':FieldArg("cssclass", ""),
                       'right_label':FieldArg("text", ''),
                       'right_class':FieldArg("cssclass", ''),
                       'right_style':FieldArg("cssstyle", ''),
                       'option_list':FieldArgList("text")
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        selectvalue: The option selected, field name is used as the widgfield attribute
        left_label: The text displayed to the left of the dropdown
        left_class: The css class of the label to the left of the dropdown
        right_label: The text displayed to the right of the dropdown
        right_class: The css class of the label to the right of the dropdown
        select_class: The css class of the select tag
        option_list: A list of options
        error_class: css class applied to the normally hidden error paragraph
        div_class: css class applied to the div after the error paragraph
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        self[1] = tag.Part(tag_name="div")
        self[1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the drop down input form
        self[1][1] = tag.Part(tag_name="select")
        self[1][2] = tag.Part(tag_name="label", hide_if_empty=True)

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the checkbox"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if self.get_field_value('div_class'):
            self[1].update_attribs({"class":self.get_field_value('div_class')})
        if self.get_field_value('left_label'):
            self[1][0][0] = self.get_field_value('left_label')
        if self.get_field_value('left_class'):
            self[1][0].attribs = {"class": self.get_field_value('left_class')}
        if self.get_field_value('left_style'):
            self[1][0].attribs = {"style": self.get_field_value('left_style')}
        if self.get_field_value('select_class'):
            self[1][1].attribs = {"class": self.get_field_value('select_class')}
        if self.get_field_value('select_style'):
            self[1][1].attribs = {"style": self.get_field_value('select_style')}
        self[1][1].update_attribs({"name":self.get_formname('selectvalue')})
        selected_option = self.get_field_value('selectvalue')
        for index, opt in enumerate(self.get_field_value('option_list')):
            if selected_option == opt:
                self[1][1][index] = tag.Part(tag_name="option", text=opt, attribs ={"selected":"selected"})
            else:
                self[1][1][index] = tag.Part(tag_name="option", text=opt)
        if self.get_field_value('right_label'):
            self[1][2][0] = self.get_field_value('right_label')
        if self.get_field_value('right_class'):
            self[1][2].attribs = {"class": self.get_field_value('right_class')}
        if self.get_field_value('right_style'):
            self[1][2].attribs = {"style": self.get_field_value('right_style')}
        # set an id in the select for the 'label for' tag
        self[1][1].insert_id()
        # set the label 'for' attribute
        self[1][0].update_attribs({'for':self[1][1].get_id()})
        self[1][2].update_attribs({'for':self[1][1].get_id()})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <div>  <!-- with class attribute set to div_class if a class is set -->
    <label> <!-- with class set to left_class and content to left_label -->
    </label>
    <select> <!-- with class set to select_class -->
     <option> <!-- with multiple options -->
     </option>
    </select>
    <label> <!-- with class set to right_class and content to right_label -->
    </label>
  </div>
</div>"""


class SubmitDropDown1(Widget):
    """Defines a form with a drop down input field, and four hidden fields"""

    error_location = (0,0,0)

    arg_descriptions = {'label':FieldArg("text", 'Choose:'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'action_json':FieldArg("url", ''),
                        'action':FieldArg("url", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True),
                        'hidden_field2':FieldArg("text", '', valdt=True),
                        'hidden_field3':FieldArg("text", '', valdt=True),
                        'hidden_field4':FieldArg("text", '', valdt=True),
                        'button_text':FieldArg("text",'Submit'),
                        'button_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ""),
                        'select_class':FieldArg("cssclass", ""),
                        'select_style':FieldArg("cssstyle", ""),
                        'selectvalue':FieldArg("text", '', valdt=True),
                        'option_list':FieldArgList("text"),
                        'hide':FieldArg("boolean", False, jsonset=True),
                       'disabled':FieldArg("boolean", False),
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        label: The text displayed to the left of the text input field
        label_class: The css class of the label
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        action: The page ident, label, url this button links to, overridden if action_json is set.
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        button_text: The text on the button
        button_class: The class given to the button tag
        inputdiv_class: class of the div holding label, dropdown and button
        select_class: The css class of the select tag
        selectvalue: The option selected, field name is used as the widgfield attribute
        option_list: A list of options
        error_class: The class applied to the paragraph containing the error message on error.
        hide: If True, widget is hidden
        disabled: Set True if the select field and submit button are to be disabled
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # error para at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''

        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})

        # div containing label, input text and button
        self[1][0] = tag.Part(tag_name='div')
        # the label
        self[1][0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the drop down input field
        self[1][0][1] = tag.Part(tag_name="select")
        # the submit button
        self[1][0][2] = tag.Part(tag_name="button", attribs ={"type":"submit"})
        self[1][0][2][0] = "Submit"
        self._jsonurl = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        self._jsonurl = skiboot.get_url(self.get_field_value("action_json"), proj_ident=page.proj_ident)
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if not self.get_field_value("action"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = skiboot.get_url(self.get_field_value("action"),  proj_ident=page.proj_ident)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].update_attribs({"action": actionurl})
        # the div holding label, dropdown and button
        if self.get_field_value('inputdiv_class'):
            self[1][0].attribs = {"class": self.get_field_value('inputdiv_class')}
        if self.get_field_value('label'):
            self[1][0][0][0] = self.get_field_value('label')
        if self.get_field_value('label_class'):
            self[1][0][0].attribs = {"class": self.get_field_value('label_class')}
        if self.get_field_value('label_style'):
            self[1][0][0].attribs = {"style": self.get_field_value('label_style')}
        if self.get_field_value('select_class'):
            self[1][0][1].attribs = {"class": self.get_field_value('select_class')}
        if self.get_field_value('select_style'):
            self[1][0][1].attribs = {"style": self.get_field_value('select_style')}
        self[1][0][1].update_attribs({"name":self.get_formname('selectvalue')})

        if self.get_field_value('disabled'):
            self[1][0][1].update_attribs({"disabled":"disabled"})

        # set an id in the input field for the 'label for' tag
        self[1][0][1].insert_id()

        selected_option = self.get_field_value('selectvalue')
        for index, opt in enumerate(self.get_field_value('option_list')):
            if selected_option == opt:
                self[1][0][1][index] = tag.Part(tag_name="option", text=opt, attribs ={"selected":"selected"})
            else:
                self[1][0][1][index] = tag.Part(tag_name="option", text=opt)

        # set the label 'for' attribute
        self[1][0][0].update_attribs({'for':self[1][0][1].get_id()})

        # submit button
        if self.get_field_value('button_text'):
            self[1][0][2][0] = self.get_field_value('button_text')
        if self.get_field_value('button_class'):
            self[1][0][2].update_attribs({"class": self.get_field_value('button_class')})
        if self.get_field_value('disabled'):
            self[1][0][2].update_attribs({"disabled":"disabled"})

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        jscript = """  $('#{ident}').on("submit", function(e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident=self.get_id())
        if self._jsonurl:
            return jscript + self._make_fieldvalues(url=self._jsonurl)
        return jscript


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div>  <!-- div hidden when no error is displayed, with class set to error_class on error -->
    <p> <!-- error message appears in this paragraph --> </p>
  </div>
  <form role="form" method="post"> <!-- action attribute set to action field -->
    <div> <!-- class attribute set to inputdiv_class -->
      <label> <!-- with class set to label_class and content to label -->
      </label>
      <select> <!-- with class set to select_class -->
       <option> <!-- with multiple options -->
       </option>
      </select>
      <button type="submit"> <!-- with class set to button_class -->
        <!-- button_text -->
      </button>
    </div>
    <!-- hidden input fields -->                              
  </form>
</div>"""






"""A Respond object instance is placed in a Respond page,
   The instance is callable, and the respondpage calls it
   to provide the page action"""

import pprint

from .. import skiboot, tag
from ..excepts import ValidateError, ServerError, FailPage, ErrorMessage

from . import Respond


class AllowedFields(Respond):
    """
Takes a list of widgfields, any which do not appear in the
submitted form data will be set into form data with an empty value. If any form
data field appears which is NOT in this list, it will be considered unwanted data,
and will cause the project validate error page to be called, if no fields are given,
the caller must not provide any form data. It also takes a list
of allowed page idents which are the possible callers to this page.

If the calling page submits no ident, and the
alternate_ident has been set - send the call there.  If it has not, then it
calls the project validate error page.

If submit_data option is True, then received data is placed in the submit_dict dictionary
under key 'received_data' which contains a dictionary of widgfield tuples:values.
"""

    # This indicates a target page ident is required
    target_ident_required = True
    
    # This indicates if another alternate page is required
    alternate_ident_required = True

    # This indicates a list of allowed caller idents is required
    allowed_callers_required = True
    
    # The form data can be validated
    validate_option_available = True

    # The form data can be submitted
    submit_option_available = True

    # Options for the fields argument
    field_options = {'fields': True,                  # If False, no fields are expected
                     'widgfields':True,               # If True, fields are widgfields, if False, can be other constants
                     'fields_optional': True,        # if fields is True, then False here means fields must be supplied
                     'widgfield_values':False,        # If True the field values are widgfields
                     'field_values': False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        "Gets the target page, filling in the form data"
        if caller_page is None:
            if self.alternate_ident:
                return self.get_alternate_page(skicall, {}, None, ident_list, proj_ident, rawformdata)
            else:
                if skiboot.get_debug():
                    responder_ident = ident_list[-1]
                    message='No caller page, and no alternate ident set in AllowedFields responder %s,%s' % responder_ident
                    raise ValidateError(message)
                else:
                    raise ValidateError(message='No caller page, and no alternate ident set in AllowedFields responder')
        self._check_allowed_callers(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # previous caller is allowed, now check the form data

        if not self.fields:
            if form_data :
                # if no fields specified, there should be no form data
                if skiboot.get_debug():
                    responder_ident = ident_list[-1]
                    message='No form data expected by AllowedFields responder %s,%s' % responder_ident
                    raise ValidateError(message)
                else:
                    raise ValidateError(message="No form data expected by AllowedFields responder")
            else:
                # no form_data received, no fields to check, go to target page or submit_data
                if self.submit_option:
                    try:
                        self._submit_data(ident_list, skicall)
                    except FailPage as e:
                        # raises a PageError exception
                        self.raise_error_page([e.errormessage], skicall, {}, caller_page, ident_list, proj_ident, rawformdata)
                return self.get_target_page(skicall, {}, caller_page, ident_list, proj_ident, rawformdata)

        # generate new form data
        new_form_data = {}
        # populate new form data with received form data
        for field in form_data:
            if field not in self.fields:
                # any form data field not in those specified by self.fields, causes an error
                if skiboot.get_debug():
                    responder_ident = ident_list[-1]
                    message='Recieved data in a field not specified in the AllowedFields responder %s,%s' % responder_ident
                    raise ValidateError(message)
                else:
                    raise ValidateError(message="Recieved data in a field not specified in the AllowedFields responder")
            new_form_data[field] = form_data[field]

        for field in self.fields:
            # add field with empty value if the field has not been submitted
            if field not in new_form_data:
                new_form_data[field] = ''
                
        if self.validate_option:
            validated_form_data = self._validate_fields(skicall, new_form_data, caller_page, ident_list, proj_ident, rawformdata)
        else:
            validated_form_data = new_form_data


        # call user submit_data
        if self.submit_option:
            received_data = {}
            for field in self.fields:
                formvalue =  validated_form_data[field]
                if isinstance(formvalue, list) or isinstance(formvalue, dict):
                    received_data[field.to_tuple_no_i()] = formvalue.copy()
                else:
                    received_data[field.to_tuple_no_i()] = formvalue

            skicall.submit_dict['received_data'] = received_data

            try:
                self._submit_data(ident_list, skicall)
            except FailPage as e:
                # raises a PageError exception
                self.raise_error_page([e.errormessage], skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)
                
        # return the target page
        return self.get_target_page(skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)


class StoreData(Respond):
    """For widgfields set here, takes the data from the submitted form, and stores them in the dictionary
       call_data with keys equal to widgfield tuples submitting the data. If widgfields specified do not appear in the form
       they will be assumed to be empty strings"""

    # This indicates a target page ident is required
    target_ident_required = True
    
    # The form data can be validated
    validate_option_available = True

    # The form data can be submitted
    submit_option_available = True

    field_options = {'fields': True,                  # If False, no fields are expected
                     'widgfields':True,               # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,        # if fields is True, then False here means fields must be supplied
                     'field_values': False,           # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):

        if self.validate_option:
            if caller_page is None:
                # raises a PageError exception
                self.raise_error_page([], skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)
            validated_form_data = self._validate_fields(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        else:
            validated_form_data = form_data

        for field in self.fields:
            if field not in validated_form_data:
                skicall.call_data[field.to_tuple_no_i()] = ''
                continue
            formvalue =  validated_form_data[field]
            if isinstance(formvalue, list) or isinstance(formvalue, dict):
                skicall.call_data[field.to_tuple_no_i()] = formvalue.copy()
            else:
                skicall.call_data[field.to_tuple_no_i()] = formvalue

        if self.submit_option:
            if caller_page is None:
                caller_ident = None
            else:
                caller_ident = caller_page.ident

            try:
                self._submit_data(ident_list, skicall)
            except FailPage as e:
                # raises a PageError exception
                self.raise_error_page([e.errormessage], skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)
 
        return self.get_target_page(skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)


class StoreDataKeyed(Respond):
    """For widgfields set here, takes the data from the submitted form, and stores them in the dictionary
       call_data.  The dictionary keys used to store these are the values given to the widgfields here."""

    # This indicates a target page ident is required
    target_ident_required = True

    # The form data can be validated
    validate_option_available = True

    # The form data can be submitted
    submit_option_available = True

    field_options = {'fields': True,                  # If False, no fields are expected
                     'widgfields':True,               # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,        # if fields is True, then False here means fields must be supplied
                     'field_values': True,            # if True, field values are used
                     'empty_values_allowed':False,    # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
    
        if self.validate_option:
            if caller_page is None:
                # raises a PageError exception
                self.raise_error_page([], skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)
            validated_form_data = self._validate_fields(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        else:
            validated_form_data = form_data

        for field, key in self.fields.items():
            if field not in validated_form_data:
                skicall.call_data[key] = ''
                continue
            formvalue =  validated_form_data[field]
            if isinstance(formvalue, list) or isinstance(formvalue, dict):
                skicall.call_data[key] = formvalue.copy()
            else:
                skicall.call_data[key] = formvalue

        if self.submit_option:
            if caller_page is None:
                caller_ident = None
            else:
                caller_ident = caller_page.ident

            try:
                self._submit_data(ident_list, skicall)
            except FailPage as e:
                # raises a PageError exception
                self.raise_error_page([e.errormessage], skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)

        return self.get_target_page(skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)


class AllowStoreKeyed(Respond):
    """
Accepts only allowed widgfields, validate submitted data
and store in call_data.

If the calling page submits no data at all, not even an ident, then if the
alternate_ident has been set - send the call there.  If it has not, then it
calls the project validate error page.

The submitted values are placed in the call_data dictionary,
with keys equal to the field values set here.

"""

    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates if another alternate page is required
    alternate_ident_required = True

    # This indicates a list of allowed caller idents is required
    allowed_callers_required = True
    
    # The form data can be validated
    validate_option_available = True

    # The form data can be submitted
    submit_option_available = True

    # Options for the fields argument
    field_options = {'fields': True,                  # If False, no fields are expected
                     'widgfields':True,               # If True, fields are widgfields, if False, can be other constants
                     'fields_optional': False,        # if fields is True, then False here means fields must be supplied
                     'widgfield_values':False,        # If True the field values are widgfields
                     'field_values': True,            # if True, field values are used
                     'empty_values_allowed':False,    # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        "Gets the target page, filling in the form data"

        if caller_page is None:
            if self.alternate_ident:
                return self.get_alternate_page(skicall, {}, None, ident_list, proj_ident, rawformdata)
            else:
                raise ValidateError(message='No caller page, and no alternate ident set in AllowStoreKeyed responder')
        self._check_allowed_callers(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # previous caller is allowed
        
        if not self.fields:
            if form_data :
                # if no fields specified, there should be no form data
                raise ValidateError(message="No form data expected by AllowedStoreKeyed responder")
            else:
                # no form_data received, no fields to check, go to target page or submit_data
                if self.submit_option:

                    try:
                        self._submit_data(ident_list, skicall)
                    except FailPage as e:
                        # raises a PageError exception
                        self.raise_error_page([e.errormessage], skicall, {}, caller_page, ident_list, proj_ident, rawformdata)
                return self.get_target_page(skicall, {}, caller_page, ident_list, proj_ident, rawformdata)
        
        # check fields received are specified in this responder
        for field in form_data:
            if field not in self.fields:
                # any form data field not in those specified by self.fields, causes an error
                raise ValidateError(message="Recieved data in a field not specified in the AllowStoreKeyed responder")

        if self.validate_option:
            validated_form_data = self._validate_fields(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        else:            
            # no validation
            validated_form_data = form_data


        for field, key in self.fields.items():
            if field not in validated_form_data:
                skicall.call_data[key] = ''
                continue
            formvalue =  validated_form_data[field]
            if isinstance(formvalue, list) or isinstance(formvalue, dict):
                skicall.call_data[key] = formvalue.copy()
            else:
                skicall.call_data[key] = formvalue

        # call user submit_data
        if self.submit_option:

            try:
                self._submit_data(ident_list, skicall)
            except FailPage as e:
                # raises a PageError exception
                self.raise_error_page([e.errormessage], skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)
        return self.get_target_page(skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)
        
        
class AllowStore(Respond):
    """For widgfields set here, takes the data from the submitted form, and stores them in the dictionary
       call_data with keys equal to widgfield tuples submitting the data"""

    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates if another alternate page is required
    alternate_ident_required = True
    
    # This indicates a list of allowed caller idents is required
    allowed_callers_required = True
    
    # The form data can be validated
    validate_option_available = True
    
    # The form data can be submitted
    submit_option_available = True

    field_options = {'fields': True,                  # If False, no fields are expected
                     'widgfields':True,               # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,        # if fields is True, then False here means fields must be supplied
                     'field_values': False,           # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):

   
        if caller_page is None:
            if self.alternate_ident:
                return self.get_alternate_page(skicall, {}, None, ident_list, proj_ident, rawformdata)
            else:
                if skiboot.get_debug():
                    responder_ident = ident_list[-1]
                    message='No caller page, and no alternate ident set in AllowStore responder %s,%s' % responder_ident
                    raise ValidateError(message)
                else:
                    raise ValidateError(message='No caller page, and no alternate ident set in AllowStore responder')
        self._check_allowed_callers(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # previous caller is allowed
        
        if not self.fields:
            if form_data :
                # if no fields specified, there should be no form data
                if skiboot.get_debug():
                    responder_ident = ident_list[-1]
                    message='No form data expected by AllowStore responder %s,%s' % responder_ident
                    raise ValidateError(message)
                else:
                    raise ValidateError(message="No form data expected by AllowStore responder")
            else:
                # no form_data received, no fields to check, go to target page or submit_data
                if self.submit_option:

                    try:
                        self._submit_data(ident_list, skicall)
                    except FailPage as e:
                        # raises a PageError exception
                        self.raise_error_page([e.errormessage], skicall, {}, caller_page, ident_list, proj_ident, rawformdata)
                return self.get_target_page(skicall, {}, caller_page, ident_list, proj_ident, rawformdata)
        
        # check fields received are specified in this responder
        for field in form_data:
            if field not in self.fields:
                # any form data field not in those specified by self.fields, causes an error
                if skiboot.get_debug():
                    responder_ident = ident_list[-1]
                    message='Recieved data in a field not specified in the AllowStore responder %s,%s' % responder_ident
                    raise ValidateError(message)
                else:
                    raise ValidateError(message="Recieved data in a field not specified in the AllowStore responder")

        if self.validate_option:
            validated_form_data = self._validate_fields(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        else:            
            # no validation
            validated_form_data = form_data

        for field in self.fields:
            if field not in validated_form_data:
                skicall.call_data[field.to_tuple_no_i()] = ''
                continue
            formvalue =  validated_form_data[field]
            if isinstance(formvalue, list) or isinstance(formvalue, dict):
                skicall.call_data[field.to_tuple_no_i()] = formvalue.copy()
            else:
                skicall.call_data[field.to_tuple_no_i()] = formvalue
                    
        # call user submit_data
        if self.submit_option:

            try:
                self._submit_data(ident_list, skicall)
            except FailPage as e:
                # raises a PageError exception
                self.raise_error_page([e.errormessage], skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)
 
        return self.get_target_page(skicall, validated_form_data, caller_page, ident_list, proj_ident, rawformdata)


class PrettyFormData(Respond):
    """Places a pretty print string of received widgfield:value's form data into submit_dict with key 'form_data'
       and pretty print of received raw data into submit_dict with key 'raw_data' and calls user submit_data function.
       Used for diagnostics. If no data available, the values are empty strings."""


    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,               # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': True,        # if fields is True, then False here means fields must be supplied
                     'field_values': False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):

        if form_data:
            new_dict = { key.to_tuple_no_i():val for key, val in form_data.items() }
            skicall.submit_dict['form_data'] = pprint.pformat(new_dict)
        else:
            skicall.submit_dict['form_data'] = ''


        # rawformdata is a FieldStorage object
        new_dict = {}
        for field in rawformdata.keys():
            value = rawformdata.getlist(field)
            if len(value) == 1:
                new_dict[field] = value[0]
            else:
                new_dict[field] = value

        if new_dict:
            skicall.submit_dict['raw_data'] = pprint.pformat(new_dict)
        else:
            skicall.submit_dict['raw_data'] = ''

        try:
            self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # so all ok, get the target page
        return self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)



class Accept(Respond):
    """places received raw data into submit_dict under key 'received_data'.
       Required when received data is not in widgfield format and without ident information"""


    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,               # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': True,        # if fields is True, then False here means fields must be supplied
                     'field_values': False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):



        # rawformdata is a FieldStorage object

        received_data = {}
        for field in rawformdata.keys():
            value = rawformdata.getlist(field)
            if len(value) == 1:
                received_data[field] = value[0]
            else:
                received_data[field] = value

        skicall.submit_dict['received_data'] = received_data

        try:
            self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        
        return self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)




class AllowedAccept(Respond):
    """places received data into submit_dict under key 'received_data', unlike Accept, the caller page must provide its ident
       otherwise alternate page is called, received data keys are widgfield tuples"""


    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates if another alternate page is required
    alternate_ident_required = True

    # This indicates a list of allowed caller idents is required
    allowed_callers_required = True

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,               # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': True,        # if fields is True, then False here means fields must be supplied
                     'field_values': False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):

        if caller_page is None:
            if self.alternate_ident:
                return self.get_alternate_page(skicall, {}, None, ident_list, proj_ident, rawformdata)
            else:
                if skiboot.get_debug():
                    responder_ident = ident_list[-1]
                    message='No caller page, and no alternate ident set in AllowedFields responder %s,%s' % responder_ident
                    raise ValidateError(message)
                else:
                    raise ValidateError(message='No caller page, and no alternate ident set in AllowedAccept responder')
        self._check_allowed_callers(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # previous caller is allowed, now store the received form data


        received_data = {}
        if form_data:
            for key, value in form_data.items():
                received_data[key.to_tuple_no_i()] = value

        skicall.submit_dict['received_data'] = received_data

        try:
            self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        
        return self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)



class PageData(Respond):
    """Places widgfield values into page_data"""

    # This indicates a target page ident is required
    target_ident_required = True

    # Options for the fields argument
    field_options = {'fields': True,                  # If False, no fields are expected
                     'widgfields':True,               # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,        # if fields is True, then False here means fields must be supplied
                     'field_values': True,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        "Places given widgfields and values into page_data"
        for field, value in self.fields.items():
            str_field = str(field)
            if ':' in str_field: 
                # a widgfield with widget:field component
                skicall.page_data[field.to_tuple_no_i()] = value
            else:
                # widgfield is actually a string, such as 'show_error'
                skicall.page_data[str_field] = value
        return self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)






import os, sys, re, collections, random


from skipole import WSGIApplication, FailPage, GoTo, ValidateError, ServerError, use_submit_list, set_debug, skilift
from skipole.skilift.fromjson import get_defaults_from_file

import skiadminpackages

PROJECT = 'skiadmin'

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')

# dictionary of session keys, and values being a dictionary of items to carry over as session data
_SESSION_DATA = collections.OrderedDict()

# A key will be created for _SESSION_DATA with a random number and an incrementing number stored in
# this global variable 
_IDENT_DATA = 0



def start_call(called_ident, skicall):
    "Checks initial incoming call parameters, and using ident_data, retrieves session data and populates call_data"

    # initially populate call_data with some project info
    editedprojname = skicall.proj_data['editedprojname']

    projinfo = skilift.project_info(editedprojname)

    skicall.call_data = {'editedprojname':editedprojname,
                         'editedprojurl':projinfo.path,
                         'editedprojversion':projinfo.version,
                         'editedprojbrief':projinfo.brief,
                         'adminproj':skicall.project,
                         'extend_nav_buttons':[],
                         'caller_ident':skicall.caller_ident}

    if called_ident is None:
        # The call is to a url not found
        return

    # If caller_ident is not given there should be no further session data
    if not skicall.caller_ident:
        return called_ident

    # if ident_data is given, then session data can be found in _SESSION_DATA[ident_data]
    # so this is added to call_data

    ident_data = skicall.ident_data

    if (ident_data) and (ident_data in _SESSION_DATA):
        # update call_data with session_data
        skicall.call_data.update(_SESSION_DATA[ident_data])

    return called_ident


# skicall.submit_list defines packages, module, function to call
@use_submit_list
def submit_data(skicall):
    """The decorator calls the appropriate submit_data function, using skicall.submit_list
       if skicall.submit_list is invalid, then this function raises ServerError"""
    raise ServerError("skicall.submit_list invalid for responder %s,%s" % skicall.ident_list[-1])


def end_call(page_ident, page_type, skicall):
    """Sets navigation menus, and stores session data under a random generated key in _SESSION_DATA
       and send the key as ident_data.
       Also limits the length of _SESSION_DATA by popping the oldest member"""

    global _SESSION_DATA, _IDENT_DATA

    # do not include session data or navigation for these types of pages
    if page_type in ('FilePage', 'CSS'):
        return

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']

    projname, identnum = page_ident

    # do not include session data if target page is in another project
    if projname != PROJECT:
        return

    # do not include session or navigation if this is the responder map page
    if identnum == 26010:
        return

    if page_type == "TemplatePage":
        # header information
        if ("adminhead","page_head","large_text") not in page_data:
            page_data["adminhead","page_head","large_text"] = "Project: %s version: %s" % (editedprojname, call_data['editedprojversion'])
        # fill in navigation information
        set_navigation(identnum, call_data, page_data)

    # Show the status message
    if 'status' in call_data:
        page_data['foot', 'foot_status','footer_text'] = call_data['status']
        page_data[("adminhead","show_status","para_text")] = call_data['status']
        page_data[("adminhead","show_status","hide")] = False


    # store any required session data which has been set into call_data

    sent_session_data = {}

    # this is a list of items to store in session data if they are available in call_data
    session_keys = ['location', 'field_arg', 'validx', 'module', 'widgetclass', 'widget_name', 'container', 'add_to_foldernumber', 
                    'section_name', 'schange', 'page_number', 'pchange', 'folder_number', 'fchange']

    for key, val in call_data.items():
        if key in session_keys:
            sent_session_data[key] = val

    if not sent_session_data:
        return

    # store as a key:value in _SESSION_DATA, and send the key as ident_data
    # this key will be saved in the returned page, and then sent back in the next call to the server
    # on being received by the start_call function, the key will be used to read the session values
    # from _SESSION_DATA

    # generate a key, being a combination of incrementing _IDENT_DATA and a random number
    _IDENT_DATA += 1
    ident_data_key = str(_IDENT_DATA) + "a" + str(random.randrange(1000, 9999))
    _SESSION_DATA[ident_data_key] = sent_session_data
    # if length of _SESSION_DATA is longer than 50, remove old values,
    # this expires old sessions
    if len(_SESSION_DATA)>50:
        _SESSION_DATA.popitem(last=False)
    page_data['ident_data'] = ident_data_key


# As this project is not intended to run as a stand-alone service, a function
# is provided rather than an application object being immediately created.


def makeapp(projectfiles, **proj_data):
    """This particular sub project does not know where its projectfiles are, so this function
       is called to provide that info, and returns the application
       skiadmin makes use of some color values, so these are set here and passed to the project
       as proj_data"""

    if 'editedprojname' not in proj_data:
        raise ServerError("The project name being edited is required")

    # get pallet of colours from defaults.json and place in proj_data
    # these will be used to populate w3-theme-ski.css and the Colours page

    adminbackcol = get_defaults_from_file(projectfiles, PROJECT, key="backcol")
    if not adminbackcol:
        adminbackcol = "#bfb786"
    colours = get_defaults_from_file(projectfiles, PROJECT, key="colours")
    if not colours:
        adminbackcol_rgb = skiadminpackages.css_styles.hex_int(adminbackcol)
        colours = skiadminpackages.css_styles.get_colours(*adminbackcol_rgb)


    proj_data["colours"] = colours
    proj_data["adminbackcol"] = adminbackcol

    # The WSGIApplication created here is generally given a URL of "/skiadmin"
    # when added to the root project using application.add_project

    return WSGIApplication(project=PROJECT,
                           projectfiles=projectfiles,
                           proj_data=proj_data,
                           start_call=start_call,
                           submit_data=submit_data,
                           end_call=end_call)


def set_navigation(identnum, call_data, page_data):
    "Sets the navigation buttons"

    # top navigation

    page_data["adminhead","top_nav","nav_links"] = [ [3, "Root Folder", False, ''] ]

    editedprojname = call_data['editedprojname']

    item_number = None

    if 'page_number' in call_data:
        item_number = call_data['page_number']
    elif 'folder_number' in call_data:
        item_number = call_data['folder_number']

    if item_number:
        # get list of [(name,number),...] of parents
        parents = skilift.parent_list(editedprojname, item_number)
        for item in parents:
            # for each item apart from root
            if item[1]:
                edited_folder = editedprojname + '_' + str(item[1])
                page_data["adminhead","top_nav","nav_links"].append(['edit_from_top_nav', item[0], False, edited_folder])
    # Top navigation lists in reverse order as the float right keeps making the next link the rightmost
    page_data["adminhead","top_nav","nav_links"].reverse()


    # Left navigation

    # uses a headers.NavButtons1 widget, each inner list consists of

    # 0 : The url, label or ident of the target page of the link
    # 1 : The displayed text of the link
    # 2 : If True, ident_data is sent with the link even if there is no get field data
    # 3 : The get field data to send with the link 

    if identnum == 20001:
        # main site index page
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            [300, "Colours", False, ''],
                                                            ['defaults', "Defaults", False, ''],
                                                            ["about_code", "Python Code", False, '']
                                                           ]
    elif (identnum == 510) or (identnum == 70105):
        # about_skis page, about skilift
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            [300, "Colours", False, ''],
                                                            ['defaults', "Defaults", False, ''],
                                                            ["about_code", "Python Code", False, '']
                                                           ]
    elif identnum == 85001:
        # operations index page
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            [300, "Colours", False, ''],
                                                            ["about_code", "Python Code", False, '']
                                                           ]
    elif identnum == 20300:
        # colours index page
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            ['defaults', "Defaults", False, ''],
                                                            ["about_code", "Python Code", False, '']
                                                           ]

    elif identnum == 70101:
        # about code page
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            [300, "Colours", False, ''],
                                                            ['defaults', "Defaults", False, '']
                                                           ]

    elif identnum == 25001:
        # page labels
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            ['defaults', "Defaults", False, '']
                                                           ]


    elif ("left_nav","navbuttons","nav_links") not in page_data:
        # all other page have full suite of navigation buttons
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, '']
                                                           ]

    widget_info = None

    if item_number:
        item_info = skilift.item_info(editedprojname, item_number)
        if item_info:
            if (item_info.item_type == 'TemplatePage') or (item_info.item_type == 'SVG'):
                # add page name to navbuttons
                if item_info.item_type == 'TemplatePage':
                    page_data["left_nav","navbuttons","nav_links"].append(['back_to_page', item_info.name, True, ''])
                    page_data["left_nav","navbuttons","nav_links"].append(['page_head', item_info.name + ' - Head', True, ''])
                    page_data["left_nav","navbuttons","nav_links"].append(['page_body', item_info.name + ' - Body', True, ''])
                if item_info.item_type == 'SVG':
                    page_data["left_nav","navbuttons","nav_links"].append(['back_to_svgpage', item_info.name, True, ''])
                    page_data["left_nav","navbuttons","nav_links"].append(['page_svg', 'SVG', True, ''])
                if ('widget_name' in call_data) and ((item_info.item_type == 'TemplatePage') or (item_info.item_type == 'SVG')):
                    widget_info = skilift.widget_info(editedprojname, item_number, None, call_data['widget_name'])

    elif 'section_name' in call_data:
        page_data["left_nav","navbuttons","nav_links"].append(['back_to_section', call_data['section_name'], True, ''])
        if 'widget_name' in call_data:
            widget_info = skilift.widget_info(editedprojname, None, call_data['section_name'], call_data['widget_name'])

    if widget_info:
        # if widget has parent, display parent links
        display_parent(widget_info, page_data)
        page_data["left_nav","navbuttons","nav_links"].append(['retrieve_widget', widget_info.name, True, widget_info.name])
        # if widget has containers, display links to them
        if widget_info.containers:
            for cont in range(widget_info.containers):
                str_cont = str(cont)
                page_data["left_nav","navbuttons","nav_links"].append(['retrieve_container', widget_info.name + " " + str_cont, True, widget_info.name + "-" + str_cont])

    # add further buttons which may be set in call_data['extend_nav_buttons']
    if 'extend_nav_buttons' in call_data:
        if call_data['extend_nav_buttons']:
            page_data["left_nav","navbuttons","nav_links"].extend(call_data['extend_nav_buttons'])


def display_parent(widget_info, page_data):
    "Appends link to parent widget in navigation buttons"
    location = widget_info.location
    # location is (parent, container, location integers) where parent is section_name, head/body or parent widget name
    # if container is None then there is no parent widget
    if location[1] is None:
        return
    # widget is in a container
    # does the parent have parents?
    parent_name = location[0]
    parent_info = skilift.widget_info(widget_info.project, widget_info.pagenumber, widget_info.section_name, parent_name)
    # recursively display grandparents
    display_parent(parent_info, page_data)
    # display links to the parent widget
    page_data["left_nav","navbuttons","nav_links"].append(['retrieve_widget', parent_name, True, parent_name])
    page_data["left_nav","navbuttons","nav_links"].append(['retrieve_container', parent_name + " " + str(location[1]), True, parent_name + "-" + str(location[1])])




